from __future__ import annotations

import os
import platform
import shutil
import subprocess
import sys

from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from qtpy.QtCore import (
    QDir,
    QFile,
    QIODevice,
    QItemSelectionModel,
    QMimeData,
    QModelIndex,
    QSortFilterProxyModel,
    QUrl,
    Qt,
    Signal,  # pyright: ignore[reportPrivateImportUsage]
)
from qtpy.QtGui import QFont, QIcon, QImage, QPalette, QPixmap, QStandardItemModel
from qtpy.QtWidgets import (
    QAbstractItemView,
    QAction,  # pyright: ignore[reportPrivateImportUsage]
    QApplication,
    QCompleter,
    QDialog,
    QFileIconProvider,
    QFileSystemModel,  # pyright: ignore[reportPrivateImportUsage]
    QInputDialog,
    QLabel,
    QListView,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from loggerplus import RobustLogger  # pyright: ignore[reportMissingTypeStubs]
from utility.gui.qt.common.actions_dispatcher import ActionsDispatcher
from utility.gui.qt.common.tasks.actions_executor import FileActionsExecutor
from utility.gui.qt.common.tasks.task_details_dialog import TaskDetailsDialog
from utility.gui.qt.filesystem.qfileexplorer.explorer_ui import Ui_QFileExplorer
from utility.gui.qt.tools.image import IconLoader
from utility.gui.qt.widgets.itemviews.abstractview import RobustAbstractItemView

if TYPE_CHECKING:
    from types import TracebackType

    from qtpy.QtCore import QAbstractItemModel, QModelIndex, QPoint
    from qtpy.QtGui import (
        QClipboard,
        QDragEnterEvent,
        QDragMoveEvent,
        QDropEvent,
        QResizeEvent,
        _QAction,
    )
    from qtpy.QtWidgets import (
        QStatusBar,
        QWidget,
        _QMenu,
    )
    from qtpy.sip import voidptr
    from typing_extensions import (
        Literal,  # pyright: ignore[reportMissingModuleSource, reportAttributeAccessIssue]
    )


class FileSystemExplorerWidget(QMainWindow):
    file_selected: Signal = Signal(str)
    directory_changed: Signal = Signal(str)
    open_in_new_tab: Signal = Signal(str)

    def __init__(  # noqa: PLR0915
        self,
        initial_path: os.PathLike | str | None = None,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.ui: Ui_QFileExplorer = Ui_QFileExplorer()
        self.ui.setupUi(self)

        # Initialize basic attributes
        self.current_path: Path = Path.home() if initial_path is None else Path(initial_path)
        self.cut_files: list[Path] | None = None

        self.fs_model: QFileSystemModel = QFileSystemModel()
        self.fs_model.setOption(QFileSystemModel.Option.DontWatchForChanges, False)  # noqa: FBT003
        self.fs_model.setOption(QFileSystemModel.Option.DontResolveSymlinks, True)  # noqa: FBT003
        self.fs_model.setRootPath(str(self.current_path.root))

        # Setup actions and connections
        self.executor = FileActionsExecutor()
        self.dispatcher: ActionsDispatcher = ActionsDispatcher(self.fs_model, self, self.executor)
        self.icon_loader: IconLoader = IconLoader()

        self.proxy_model: QSortFilterProxyModel = QSortFilterProxyModel()
        self.proxy_model.setSourceModel(self.fs_model)
        self.completer: QCompleter = QCompleter(self)
        self.completer.setModel(self.fs_model)

        self.ui.fileSystemTreeView.setModel(self.proxy_model)
        self.ui.dynamicView.setModel(self.proxy_model)
        self.ui.dynamicView.setRootIndex(self.proxy_model.mapFromSource(self.fs_model.index(str(self.current_path))))
        self.ui.dynamicView.show()
        for view in self.ui.dynamicView.all_views():
            view.clicked.connect(self.on_file_list_view_clicked)
            view.doubleClicked.connect(self.on_item_double_clicked)
        self.ui.dynamicView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.ui.dynamicView.customContextMenuRequested.connect(self.show_context_menu)
        self.ui.dynamicView.selectionModel().selectionChanged.connect(self.update_preview)
        self.bookmarks_model: QStandardItemModel = QStandardItemModel()
        self.ui.bookmarksListView.setModel(self.bookmarks_model)
        self.ui.bookmarksListView.clicked.connect(self.on_bookmark_clicked)

        # Setup other UI components
        self.ui.fileSystemTreeView.setModel(self.fs_model)
        self.ui.fileSystemTreeView.setRootIndex(self.fs_model.index(QDir.rootPath()))
        self.ui.fileSystemTreeView.clicked.connect(self.on_sidepanel_treeview_clicked)
        self.ui.fileSystemTreeView.expanded.connect(self.on_treeview_expanded)
        self.ui.fileSystemTreeView.collapsed.connect(self.on_treeview_collapsed)
        for i in range(1, self.fs_model.columnCount()):  # Show only the Name column in the tree view
            self.ui.fileSystemTreeView.hideColumn(i)
        self.ui.fileSystemTreeView.clicked.connect(self.on_navigation_pane_clicked)
        self.ui.addressBar.refreshButton.clicked.connect(lambda: self.refresh())
        self.ui.addressBar.pathChanged.connect(self.on_address_bar_path_changed)
        self.ui.addressBar.returnPressed.connect(self.on_address_bar_return)

        self.ui.zoom_slider.valueChanged.connect(self.on_zoom_slider_changed)
        self.ui.itemCountLabel.linkActivated.connect(self.on_item_count_clicked)
        self.ui.selectedCountLabel.linkActivated.connect(self.on_selected_count_clicked)
        self.ui.freeSpaceLabel.linkActivated.connect(self.on_free_space_clicked)
        self.ui.taskStatusToggle.clicked.connect(self.toggle_task_status_widget)

        self.executor.TaskStarted.connect(self.on_task_started)
        self.executor.TaskCompleted.connect(self.on_task_completed)
        self.executor.TaskFailed.connect(self.on_task_failed)
        self.executor.TaskCancelled.connect(self.on_task_cancelled)
        self.executor.TaskProgress.connect(self.on_task_progress)
        self.executor.AllTasksCompleted.connect(self.on_all_tasks_completed)
        self.executor.ProgressUpdated.connect(self.update_progress_bar)
        self.ui.taskStatusWidget.task_clicked.connect(self.show_task_details)
        self.ui.previewWidget.setVisible(False)
        status_bar: QStatusBar | None = self.statusBar()
        assert status_bar is not None
        status_bar.addPermanentWidget(self.ui.taskStatusWidget)
        self.task_status_toggle: QPushButton = QPushButton("Task Status")
        self.task_status_toggle.setCheckable(True)
        self.task_status_toggle.setChecked(True)
        self.task_status_toggle.clicked.connect(self.toggle_task_status_widget)
        self.ui.progressBar.mousePressEvent = self.show_task_details  # pyright: ignore[reportAttributeAccessIssue]
        self.ui.progressBar.hide()
        drive_layout = self.ui.horizontalLayout_2
        for drive in QDir.drives():
            drive_path = drive.path()
            drive_button = QPushButton(drive_path)
            drive_button.clicked.connect(lambda _, path=drive_path: self.set_current_path(path))
            drive_layout.addWidget(drive_button)

        self.dispatcher.menus.actions.actionTiles.triggered.connect(lambda: self.ui.dynamicView.list_view().setViewMode(QListView.ViewMode.IconMode))
        self.dispatcher.menus.actions.actionListView.triggered.connect(lambda: self.ui.dynamicView.list_view().setViewMode(QListView.ViewMode.ListMode))
        self.dispatcher.menus.actions.actionDetailView.triggered.connect(lambda: self.ui.dynamicView.setCurrentWidget(self.ui.dynamicView.tree_view()))

        if self.ui.searchBar.isVisible():
            self.ui.searchBar.hide()
        else:
            self.ui.searchBar.show()
            self.ui.searchBar.setFocus()

        self.ui.searchBar.textChanged.connect(self.on_search_text_changed)
        self.ui.searchBar.setPlaceholderText("Search...")

        # Add ribbon toggle button
        self.toggle_ribbon_button = QPushButton("▼", self)
        self.toggle_ribbon_button.setToolTip("Toggle Ribbon Menu")
        self.toggle_ribbon_button.setMaximumWidth(20)
        self.toggle_ribbon_button.clicked.connect(self.toggle_ribbon)

        # Adjust UI sizes
        self.resize(1000, 600)
        self.ui.mainSplitter.setSizes([200, 800])
        self.ui.sidebarToolBox.setMinimumWidth(150)
        self.ui.dynamicView.setMinimumWidth(400)
        self.ui.searchAndAddressWidget.setFixedHeight(30)
        self.fs_model.directoryLoaded.connect(self.on_directory_loaded)
        self.fs_model.setReadOnly(False)
        self.fs_model.setOption(QFileSystemModel.Option.DontWatchForChanges, False)  # noqa: FBT003
        self.icon_provider: QFileIconProvider = QFileIconProvider()
        self.fs_model.setIconProvider(self.icon_provider)

        # Setup Platform-specific features
        self.setup_platform_features()

        # Apply styling
        self.apply_modern_style()
        self.apply_progress_bar_style()

        # Final UI updates
        self.update_ui()
        self.ui.addressBar.update_path(self.current_path)
        self.ui.dynamicView.setRootIndex(self.proxy_model.mapFromSource(self.fs_model.index(str(self.current_path))))

    def setup_platform_features(self):
        """Massive TODO."""
        if platform.system() == "Windows":
            try:
                import comtypes.client as cc  # pyright: ignore[reportMissingTypeStubs]

                self.shell = cc.CreateObject("Shell.Application")
            except ImportError:
                try:
                    import win32com.client

                    self.shell = win32com.client.Dispatch("Shell.Application")
                except ImportError:
                    RobustLogger().warning("Neither comtypes nor pywin32 is available. COM interfaces will not be used.")
                    self.shell = None

    def selectedFiles(self) -> list[str]:
        selected_proxy_indexes: list[QModelIndex] = self.ui.dynamicView.selectedIndexes()
        selected_source_indexes: list[QModelIndex] = [self.proxy_model.mapToSource(index) for index in selected_proxy_indexes]
        return [self.fs_model.filePath(index) for index in selected_source_indexes]

    def proxyModel(self) -> QSortFilterProxyModel:
        return self.proxy_model

    def apply_modern_style(self):
        palette: QPalette = self.palette()
        background_color: str = palette.color(QPalette.ColorRole.Base).name()
        toolbar_color: str = palette.color(QPalette.ColorRole.Midlight).name()
        border_color: str = palette.color(QPalette.ColorRole.Mid).name()
        title_color: str = palette.color(QPalette.ColorRole.Midlight).name()

        self.setStyleSheet(f"""
            QMainWindow {{
                background-color: {background_color};
            }}
            QToolBar {{
                border: none;
                background-color: {toolbar_color};
            }}
            QDockWidget {{
                border: 1px solid {border_color};
            }}
            QDockWidget::title {{
                background-color: {title_color};
                padding-left: 5px;
            }}
        """)

    def apply_progress_bar_style(self):
        palette: QPalette = self.palette()
        self.ui.progressBar.setStyleSheet(f"""
            QProgressBar {{
                border: 1px solid {palette.color(QPalette.ColorRole.Mid).name()};
                border-radius: 2px;
                text-align: center;
                font-size: 10px;
            }}
            QProgressBar::chunk {{
                background-color: {palette.color(QPalette.ColorRole.Highlight).name()};
                width: 5px;
                margin: 0.5px;
            }}
        """)

    def connect_signals(self):
        self.executor.TaskStarted.connect(self.on_task_started)
        self.executor.TaskCompleted.connect(self.on_task_completed)
        self.executor.TaskFailed.connect(self.on_task_failed)
        self.executor.TaskCancelled.connect(self.on_task_cancelled)
        self.executor.TaskProgress.connect(self.on_task_progress)
        self.executor.AllTasksCompleted.connect(self.on_all_tasks_completed)
        self.executor.ProgressUpdated.connect(self.update_progress_bar)
        self.ui.taskStatusWidget.set_dispatcher(self.executor)
        self.ui.taskStatusWidget.task_clicked.connect(self.show_task_details)
        self.ui.dynamicView.selectionModel().selectionChanged.connect(self.update_preview)

    def on_task_started(self, task_id: int):
        cast("QStatusBar", self.statusBar()).showMessage(f"Task {task_id} started")

    def on_task_completed(self, task_id: int, result: Any):
        cast("QStatusBar", self.statusBar()).showMessage(f"Task {task_id} completed")
        self.refresh()

    def on_task_failed(self, task_id: int, error: Exception):
        cast("QStatusBar", self.statusBar()).showMessage(f"Task {task_id} failed: {error}")

    def on_task_cancelled(self, task_id: int):
        cast("QStatusBar", self.statusBar()).showMessage(f"Task {task_id} cancelled")

    def on_task_progress(self, task_id: int, progress: float):
        cast("QStatusBar", self.statusBar()).showMessage(f"Task {task_id} progress: {progress:.2f}%")

    def on_all_tasks_completed(self):
        cast("QStatusBar", self.statusBar()).showMessage("All tasks completed")
        self.refresh()

    def toggle_task_status_widget(self) -> None:
        """Toggle the visibility of the task status widget."""
        if self.ui.taskStatusWidget.isVisible():
            self.ui.taskStatusWidget.hide()
            self.task_status_toggle.setChecked(False)
        else:
            self.ui.taskStatusWidget.show()
            self.task_status_toggle.setChecked(True)

    def toggle_ribbon(self):
        """Toggle the visibility of the ribbon menu."""
        if self.ui.ribbonWidget.isVisible():
            self.ui.ribbonWidget.hide()
            self.toggle_ribbon_button.setText("▼")
        else:
            self.ui.ribbonWidget.show()
            self.toggle_ribbon_button.setText("▲")

    def on_sidepanel_treeview_clicked(
        self,
        index: QModelIndex,
    ) -> None:
        path: str = self.fs_model.filePath(index)
        self.set_current_path(path)

    def on_bookmark_clicked(
        self,
        index: QModelIndex,
    ) -> None:
        path: str = self.bookmarks_model.data(index, Qt.ItemDataRole.UserRole)
        self.set_current_path(path)

    def on_zoom_slider_changed(
        self,
        value: int,
    ) -> None:
        # Implement zoom functionality
        view: QAbstractItemView | None = self.ui.dynamicView.current_view()
        assert isinstance(view, QAbstractItemView), f"View is not a QAbstractItemView, instead was a {type(view)}"
        if isinstance(view, RobustAbstractItemView):
            view.set_text_size(value)
        else:
            view.setFont(QFont(view.font().family(), value))

    def on_item_count_clicked(
        self,
        link: QLabel,
    ) -> None:
        # Handle item count label click
        total_items: int = self.fs_model.rowCount(self.ui.dynamicView.rootIndex())
        QMessageBox.information(self, "Item Count", f"Total items: {total_items}")

    def on_selected_count_clicked(
        self,
        link: QLabel,
    ) -> None:
        # Handle selected count label click
        selected_items: int = len(self.ui.dynamicView.selectedIndexes())
        QMessageBox.information(self, "Selected Count", f"Selected items: {selected_items}")

    def on_free_space_clicked(
        self,
        link: QLabel,
    ) -> None:
        # Handle free space label click
        free_space: int = shutil.disk_usage(self.current_path).free
        QMessageBox.information(self, "Free Space", f"Free space on {self.current_path}: {self.format_size(free_space)}")

    def on_navigation_pane_clicked(
        self,
        index: QModelIndex,
    ) -> None:
        path: str = self.fs_model.filePath(index)
        self.set_current_path(path)

    def on_file_list_view_clicked(
        self,
        index: QModelIndex,
    ) -> None:
        source_index: QModelIndex = self.proxy_model.mapToSource(index)
        path: str = self.fs_model.filePath(source_index)
        self.file_selected.emit(path)

    def on_item_double_clicked(
        self,
        index: QModelIndex,
    ) -> None:
        source_index: QModelIndex = self.proxy_model.mapToSource(index)
        path: str = self.fs_model.filePath(source_index)
        if self.fs_model.isDir(source_index):
            self.set_current_path(path)
        else:
            os.startfile(path)  # noqa: S606
            self.executor.queue_task("open_file", (path,))

    def on_address_bar_return(self):
        path = self.ui.addressBar.line_edit.text()
        self.set_current_path(path)

    def on_go_button_clicked(self):
        self.on_address_bar_return()

    def set_current_path(
        self,
        path: os.PathLike | str,
    ) -> None:
        path_obj: Path = Path(path)
        if path_obj.is_dir():
            self.current_path = path_obj
            source_index: QModelIndex = self.fs_model.index(str(path_obj))
            proxy_index: QModelIndex = self.proxy_model.mapFromSource(source_index)
            self.ui.dynamicView.setRootIndex(proxy_index)
            self.ui.fileSystemTreeView.setCurrentIndex(source_index)
            self.ui.addressBar.update_path(path_obj)
            self.directory_changed.emit(str(path_obj))
            self.ui.addressBar._update_history(path_obj)  # noqa: SLF001
            self.update_ui()
        else:
            # Handle invalid path
            self.ui.addressBar.update_path(self.current_path)

    def on_search_text_changed(
        self,
        text: str,
    ) -> None:
        self.proxy_model.setFilterRegularExpression(text)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        proxy_index: QModelIndex = self.proxy_model.mapFromSource(self.fs_model.index(str(self.current_path)))
        self.ui.dynamicView.setRootIndex(proxy_index)

    def focus_address_bar(self):
        self.ui.addressBar.setFocus()
        self.ui.addressBar.line_edit.selectAll()

    def connect_menu_actions(
        self,
        menu: _QMenu,
    ) -> None:
        for action in menu.actions():
            if isinstance(action, QAction):
                action.triggered.connect(lambda checked, a=action: self.execute_menu_action(a))
            elif isinstance(action, QMenu):
                self.connect_menu_actions(action)

    def execute_menu_action(
        self,
        action: _QAction,
    ):
        action_name = action.text().lower().replace(" ", "_")
        method_name = f"on_{action_name}"
        if hasattr(self, method_name):
            getattr(self, method_name)()
        else:
            print(f"Method {method_name} not implemented")

    def on_open_in_new_window(self):
        index: int = self.ui.dynamicView.currentIndex()
        path: str = self.fs_model.filePath(self.proxy_model.mapToSource(index))
        new_window = FileSystemExplorerWidget(initial_path=path)
        new_window.show()
        new_window.raise_()
        new_window.activateWindow()

    def on_open_in_new_tab(self):
        index = self.ui.dynamicView.currentIndex()
        path = self.fs_model.filePath(self.proxy_model.mapToSource(index))
        self.open_in_new_tab.emit(str(path))  # Emit signal to be handled by parent widget

    def on_properties(self):
        index = self.ui.dynamicView.currentIndex()
        path = self.fs_model.filePath(self.proxy_model.mapToSource(index))
        self.executor.queue_task("get_properties", args=(path,), callback=self.show_properties, priority=1)

    def show_properties(self, properties: dict):
        dialog = PropertiesDialog(properties, self)
        dialog.exec()

    def show_properties2(
        self,
        properties: dict,
    ) -> None:
        """Show file or directory properties."""
        property_text: str = "\n".join(f"{k}: {v}" for k, v in properties.items())
        QMessageBox.information(self, "Properties", property_text)

    def update_status_bar(self):
        selected_items: int = len(self.ui.dynamicView.selectedIndexes())
        view_model: QAbstractItemModel | None = self.ui.dynamicView.current_view().model()
        assert view_model is not None, "View model is None"

        # Map the proxy index to the source index
        proxy_root_index: QModelIndex = self.ui.dynamicView.rootIndex()
        source_root_index: QModelIndex = self.proxy_model.mapToSource(proxy_root_index)

        total_items = self.fs_model.rowCount(source_root_index)

        self.ui.itemCountLabel.setText(f"{total_items} item{'s' if total_items != 1 else ''}")
        self.ui.selectedCountLabel.setText(f"{selected_items} selected")

        root_path = self.fs_model.filePath(source_root_index)
        try:
            free_space = shutil.disk_usage(root_path).free
            self.ui.freeSpaceLabel.setText(f"Free space: {self.format_size(free_space)}")
        except OSError:
            self.ui.freeSpaceLabel.setText("Free space: Unknown")

        # Update progress bar
        completed_tasks: int = self.executor.completed_tasks
        total_tasks: int = self.executor.total_tasks
        if total_tasks > 0:
            progress = int((completed_tasks / total_tasks) * 100)
            self.ui.progressBar.setValue(progress)
            self.ui.progressBar.setFormat(f"{completed_tasks}/{total_tasks}")
            self.ui.progressBar.setVisible(True)
        else:
            self.ui.progressBar.setVisible(False)

    def format_size(self, size: int) -> str:
        readable_size: float = size
        for unit in ["B", "KB", "MB", "GB", "TB"]:
            if readable_size < 1024.0:  # noqa: PLR2004
                return f"{readable_size:.1f} {unit}"
            readable_size /= 1024.0
        return f"{readable_size:.1f} PB"

    def new_window(self):
        new_window = FileSystemExplorerWidget()
        new_window.show()

    def open_windows_terminal(self):
        import shlex

        command: str = f'start cmd /K "cd /d {self.current_path}"'
        args: list[str] = shlex.split(command)
        subprocess.Popen(args)  # noqa: S603

    def cut(self):
        self.copy(cut=True)

    def copy(self, cut: bool = False):  # noqa: FBT001, FBT002
        selected_indexes: list[QModelIndex] = self.ui.dynamicView.selectedIndexes()
        if selected_indexes:
            source_paths: list[Path] = [Path(self.fs_model.filePath(index)) for index in selected_indexes]  # pyright: ignore[reportArgumentType]
            mime_data: QMimeData = QMimeData()
            urls: list[QUrl] = [QUrl.fromLocalFile(str(path)) for path in source_paths]
            mime_data.setUrls(urls)
            cb: QClipboard | None = None
            QApplication.clipboard().setMimeData(mime_data)

            operation: Literal["cut"] | Literal["copy"] = "cut" if cut else "copy"
            self.executor.queue_task(
                lambda _: setattr(self, "to_cut", source_paths if cut else None),
                args=(source_paths,),
            )

    def paste(self):
        mime_data: QMimeData | None = QApplication.clipboard().mimeData()
        if mime_data and mime_data.hasUrls():
            source_paths: list[Path] = [Path(url.toLocalFile()) for url in mime_data.urls()]
            destination_path: Path = self.current_path

            def on_complete(_):
                self.refresh()
                if hasattr(self, "to_cut") and self.to_cut == source_paths:
                    self.executor.queue_task("delete_items", args=(self.to_cut,), on_complete=lambda _: setattr(self, "to_cut", None))

            self.executor.queue_task("paste", args=(source_paths, destination_path), on_complete=on_complete)

    def delete_items(self):
        selected_indexes: list[QModelIndex] = self.ui.dynamicView.selectedIndexes()
        if selected_indexes:
            paths: list[Path] = [Path(self.fs_model.filePath(index)) for index in selected_indexes]
            reply: QMessageBox.StandardButton = QMessageBox.question(
                self,
                "Confirm Delete",
                f"Are you sure you want to delete {len(paths)} item(s)?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,  # pyright: ignore[reportOperatorType, reportArgumentType]
                QMessageBox.StandardButton.No,
            )
            if reply == QMessageBox.StandardButton.Yes:
                task_id = self.executor.queue_task("delete_items", args=(paths,), priority=1, description=f"Deleting {len(paths)} item(s)")
                self.update_ui()
                self.show_task_details(task_id)

    def show_error_message(self, error: Exception):
        QMessageBox.critical(self, "Error", str(error))

    def update_ui(self):
        self.update_status_bar()
        # self.update_task_status_widget()

    def update_task_status_widget(self):
        self.ui.taskStatusWidget.update_all_tasks()

    def show_task_details(self, task_id: int):
        task_details_dialog = TaskDetailsDialog(self.executor, task_id, self)
        task_details_dialog.exec()

    def update_progress(
        self,
        value: int,
    ):
        self.ui.progressBar.setValue(value)
        if value >= 100:  # noqa: PLR2004
            self.ui.progressBar.setVisible(False)
        elif not self.ui.progressBar.isVisible():
            self.ui.progressBar.setVisible(True)

    def select_all(self):
        self.ui.dynamicView.selectAll()

    def invert_selection(self):
        selection_model: QItemSelectionModel = self.ui.dynamicView.selectionModel()
        assert isinstance(selection_model, QItemSelectionModel), f"Selection model is not a QItemSelectionModel, instead was a {type(selection_model)}"
        for i in range(self.fs_model.rowCount(self.ui.dynamicView.rootIndex())):  # pyright: ignore[reportArgumentType]
            index: QModelIndex = self.fs_model.index(i, 0, self.ui.dynamicView.rootIndex())  # pyright: ignore[reportArgumentType]
            selection_model.select(index, QItemSelectionModel.SelectionFlag.Toggle)  # pyright: ignore[reportCallIssue, reportArgumentType]

    def select_none(self):
        self.ui.dynamicView.clearSelection()

    def refresh(self):
        current_index: QModelIndex = self.fs_model.index(str(self.current_path))
        self.fs_model.setRootPath(str(self.current_path))
        self.ui.dynamicView.setRootIndex(current_index)
        self.update_ui()

    def on_address_bar_path_changed(
        self,
        path: Path,
    ):
        self.change_path(path)

    def on_directory_loaded(
        self,
        path: str,
    ):
        loaded_path: Path = Path(path)
        if loaded_path == self.current_path:
            self.ui.addressBar.update_path(loaded_path)
        for i in range(self.fs_model.rowCount(self.fs_model.index(path))):
            child_index = self.fs_model.index(i, 0, self.fs_model.index(path))
            file_path = self.fs_model.filePath(child_index)

    def on_treeview_expanded(self, index: QModelIndex):
        self.fs_model.fetchMore(index)

    def on_treeview_collapsed(self, index: QModelIndex): ...

    def change_path(
        self,
        path: Path,
    ):
        self.current_path = path
        source_index: QModelIndex = self.fs_model.index(str(path))
        proxy_index: QModelIndex = self.proxy_model.mapFromSource(source_index)
        self.ui.dynamicView.setRootIndex(proxy_index)
        self.ui.fileSystemTreeView.setCurrentIndex(source_index)
        self.ui.addressBar.update_path(path)

    def on_view_clicked(
        self,
        index: QModelIndex,
    ):
        source_index = self.proxy_model.mapToSource(index)
        path = Path(self.fs_model.filePath(source_index))
        self.change_path(path)

    def resizeEvent(
        self,
        event: QResizeEvent,
    ):
        super().resizeEvent(event)

    def show_context_menu(
        self,
        pos: QPoint,
    ):
        current_view: QWidget | None = self.ui.dynamicView.current_view()
        assert isinstance(current_view, QAbstractItemView), f"Current view is not a QListView, instead was a {type(current_view).__name__}"
        index = current_view.indexAt(pos)
        if not index.isValid():
            current_view.clearSelection()
        menu: QMenu = self.dispatcher.get_context_menu(current_view, pos)
        menu.exec(current_view.viewport().mapToGlobal(pos))

    def create_new_folder(self):
        folder_name, ok = QInputDialog.getText(self, "New Folder", "Enter folder name:")
        if ok and folder_name:
            new_folder_path = Path(self.current_path, folder_name)
            try:
                new_folder_path.mkdir(parents=True, exist_ok=False)
                self.refresh()
            except FileExistsError:
                QMessageBox.warning(self, "Error", f"A folder named '{folder_name}' already exists.")
            except PermissionError:
                QMessageBox.warning(self, "Error", f"Permission denied. Unable to create folder '{folder_name}'.")
            except OSError as e:
                QMessageBox.warning(self, "Error", f"Failed to create folder '{folder_name}'. Error: {e!s}")

    def create_new_file(self):
        file_name, ok = QInputDialog.getText(self, "New File", "Enter file name:")
        if ok and file_name:
            new_file_path = QDir(str(self.current_path)).filePath(file_name)
            file = QFile(new_file_path)
            if file.open(QIODevice.OpenModeFlag.WriteOnly):
                file.close()
                self.fs_model.setRootPath(str(self.current_path))  # Refresh the view
            else:
                QMessageBox.warning(self, "Error", "Failed to create file.")

    def focus_search_bar(self):
        self.ui.searchBar.setFocus()
        self.ui.searchBar.selectAll()

    def dragEnterEvent(
        self,
        event: QDragEnterEvent,
    ):
        my_mimedata: QMimeData | None = event.mimeData()
        assert my_mimedata is not None
        if my_mimedata.hasUrls():
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(
        self,
        event: QDragMoveEvent,
    ):
        my_mimedata: QMimeData | None = event.mimeData()
        assert my_mimedata is not None
        if my_mimedata.hasUrls():
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(
        self,
        event: QDropEvent,
    ):
        my_mimedata: QMimeData | None = event.mimeData()
        assert my_mimedata is not None
        if my_mimedata.hasUrls():
            event.setDropAction(Qt.DropAction.CopyAction)
            event.accept()
            urls = my_mimedata.urls()
            paths = [Path(url.toLocalFile()) for url in urls]
            for path in paths:
                if path.is_dir():
                    self.executor.queue_task("copy_directory", args=(path, self.current_path), callback=self.refresh)
                else:
                    self.executor.queue_task("copy_file", args=(path, self.current_path), callback=self.refresh)
            self.statusBar().showMessage(f"Copying {len(paths)} item(s)...")
        else:
            event.ignore()

    def create_mime_data(self, paths: list[Path]) -> QMimeData:
        mime_data = QMimeData()
        urls: list[QUrl] = [QUrl.fromLocalFile(str(path)) for path in paths]
        mime_data.setUrls(urls)
        return mime_data

    def update_preview(self):
        selected_indexes: list[QModelIndex] = self.ui.dynamicView.selectedIndexes()
        if selected_indexes:
            file_path = Path(self.fs_model.filePath(selected_indexes[0]))
            if file_path.is_file():
                if file_path.suffix.lower() in [".jpg", ".jpeg", ".png", ".gif", ".bmp"]:
                    self.icon_loader.load_thumbnail(str(file_path))
                elif file_path.suffix.lower() in [".txt", ".py", ".cpp", ".h", ".html", ".css", ".js"]:
                    self.load_text_preview(file_path)
                elif file_path.suffix.lower() in [".pdf"]:
                    self.load_pdf_preview(file_path)
                else:
                    self.icon_loader.load_icon(str(file_path))
                self.ui.previewWidget.setVisible(True)
            else:
                self.load_folder_preview(file_path)
                self.ui.previewWidget.setVisible(True)
        else:
            self.ui.previewWidget.setVisible(False)

    def load_text_preview(
        self,
        file_path: Path,
    ):
        try:
            with open(file_path, encoding="utf-8") as f:
                content = f.read(1000)  # Read first 1000 characters
            self.ui.previewWidget.set_text(content)
        except Exception as e:
            self.ui.previewWidget.set_text(f"Error loading text preview: {e!s}")

    def load_pdf_preview(
        self,
        file_path: Path,
    ):
        try:
            import fitz  # PyMuPDF

            doc = fitz.open(file_path)
            page = doc.load_page(0)  # Load the first page
            pix = page.get_pixmap()
            img = QImage(pix.samples, pix.width, pix.height, pix.stride, QImage.Format_RGB888)
            self.ui.previewWidget.set_image(img)
        except ImportError:
            self.ui.previewWidget.set_text("PDF preview requires PyMuPDF library")
        except Exception as e:
            self.ui.previewWidget.set_text(f"Error loading PDF preview: {e!s}")

    def load_folder_preview(self, folder_path: Path):
        try:
            items = list(folder_path.iterdir())
            preview_text = f"Folder: {folder_path.name}\n"
            preview_text += f"Items: {len(items)}\n"
            preview_text += f"Size: {self.format_size(sum(f.stat().st_size for f in items if f.is_file()))}\n"
            preview_text += "\nContents:\n"
            for item in items[:10]:  # Show first 10 items
                preview_text += f"- {item.name}\n"
            if len(items) > 10:
                preview_text += "..."
            self.ui.previewWidget.set_text(preview_text)
        except Exception as e:
            self.ui.previewWidget.set_text(f"Error loading folder preview: {e!s}")

    def on_icon_loaded(
        self,
        file_path: str,
        icon: QIcon,
    ):
        self.ui.previewWidget.set_icon(icon)
        self.fs_model.setData(self.fs_model.index(file_path), icon, Qt.ItemDataRole.DecorationRole)

    def on_thumbnail_loaded(
        self,
        file_path: str,
        image: QImage,
    ):
        val: voidptr | None = image.bits()
        if val is None:
            return
        if not isinstance(val, bytes):
            val = bytes(val.asarray())
        self.ui.previewWidget.set_image(val, image.width(), image.height())
        icon = QIcon(QPixmap.fromImage(image))
        self.fs_model.setData(self.fs_model.index(file_path), icon, Qt.ItemDataRole.DecorationRole)

    def update_progress_bar(
        self,
        completed_tasks: int,
        total_tasks: int,
    ) -> None:
        """Update the progress bar with the current task completion status."""
        progress = (completed_tasks / total_tasks) * 100 if total_tasks > 0 else 0
        self.ui.progressBar.setValue(int(progress))
        self.ui.progressBar.setFormat(f"{completed_tasks}/{total_tasks} tasks completed")
        self.ui.progressBar.setStyleSheet("""
            QProgressBar {
                border: 1px solid grey;
                border-radius: 5px;
                text-align: center;
            }
            QProgressBar::chunk {
                background-color: #4CAF50;
                width: 10px;
                margin: 0.5px;
            }
        """)
        self.ui.progressBar.show()


if __name__ == "__main__":

    def onAppCrash(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
    ) -> None:
        error_message = f"An unexpected error occurred:\n{exc_type.__name__}: {exc_value}"
        RobustLogger().exception(error_message, exc_info=(exc_type, exc_value, exc_traceback))
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = onAppCrash
    app = QApplication(sys.argv)
    main_window = QMainWindow()
    file_explorer = FileSystemExplorerWidget()
    main_window.setCentralWidget(file_explorer)
    main_window.setWindowTitle("File System Explorer")
    main_window.resize(1000, 600)
    main_window.show()
    sys.exit(app.exec())


class PropertiesDialog(QDialog):
    def __init__(
        self,
        properties: dict,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.setWindowTitle("Properties")
        layout = QVBoxLayout()
        self.setLayout(layout)

        for key, value in properties.items():
            label = QLabel(f"{key}: {value}")
            layout.addWidget(label)

        close_button = QPushButton("Close")
        close_button.clicked.connect(lambda: self.accept())
        layout.addWidget(close_button)



from __future__ import annotations

import os
import re
import sys

from abc import abstractmethod
from enum import Flag, IntEnum, IntFlag
from typing import TYPE_CHECKING

from qtpy.QtCore import QObject, QUrl, Qt, Signal  # pyright: ignore[reportPrivateImportUsage]
from qtpy.QtGui import QColor, QFont
from qtpy.QtWidgets import QFileDialog, QMessageBox

from utility.gui.qt.adapters.kernel.qplatformdialoghelper.qfiledialogoptions import (
    QFileDialogOptions,
)

if TYPE_CHECKING:
    from typing_extensions import Self


class StyleHint(IntEnum):
    DialogIsQtWindow = 0


class DialogCode(IntEnum):
    Rejected = 0
    Accepted = 1


class StandardButton(IntFlag):
    NoButton = 0x00000000
    Ok = 0x00000400
    Save = 0x00000800
    SaveAll = 0x00001000
    Open = 0x00002000
    Yes = 0x00004000
    YesToAll = 0x00008000
    No = 0x00010000
    NoToAll = 0x00020000
    Abort = 0x00040000
    Retry = 0x00080000
    Ignore = 0x00100000
    Close = 0x00200000
    Cancel = 0x00400000
    Discard = 0x00800000
    Help = 0x01000000
    Apply = 0x02000000
    Reset = 0x04000000
    RestoreDefaults = 0x08000000


class ButtonRole(IntFlag):
    InvalidRole = -1
    AcceptRole = 0
    RejectRole = 1
    DestructiveRole = 2
    ActionRole = 3
    HelpRole = 4
    YesRole = 5
    NoRole = 6
    ResetRole = 7
    ApplyRole = 8
    RoleMask = 0x0FFFFFFF
    AlternateRole = 0x10000000
    Stretch = 0x20000000
    Reverse = 0x40000000
    EOL = InvalidRole


class ButtonLayout(IntEnum):
    UnknownLayout = -1
    WinLayout = 0
    MacLayout = 1
    KdeLayout = 2
    GnomeLayout = 3
    AndroidLayout = 4


BUTTON_ROLE_LAYOUTS = (
    (
        (
            ButtonRole.ResetRole,
            ButtonRole.Stretch,
            ButtonRole.YesRole,
            ButtonRole.AcceptRole,
            ButtonRole.AlternateRole,
            ButtonRole.DestructiveRole,
            ButtonRole.NoRole,
            ButtonRole.ActionRole,
            ButtonRole.RejectRole,
            ButtonRole.ApplyRole,
            ButtonRole.HelpRole,
            ButtonRole.EOL,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.HelpRole,
            ButtonRole.ResetRole,
            ButtonRole.ApplyRole,
            ButtonRole.ActionRole,
            ButtonRole.Stretch,
            ButtonRole.DestructiveRole | ButtonRole.Reverse,
            ButtonRole.AlternateRole | ButtonRole.Reverse,
            ButtonRole.RejectRole | ButtonRole.Reverse,
            ButtonRole.AcceptRole | ButtonRole.Reverse,
            ButtonRole.NoRole | ButtonRole.Reverse,
            ButtonRole.YesRole | ButtonRole.Reverse,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.HelpRole,
            ButtonRole.ResetRole,
            ButtonRole.Stretch,
            ButtonRole.YesRole,
            ButtonRole.NoRole,
            ButtonRole.ActionRole,
            ButtonRole.AcceptRole,
            ButtonRole.AlternateRole,
            ButtonRole.ApplyRole,
            ButtonRole.DestructiveRole,
            ButtonRole.RejectRole,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.HelpRole,
            ButtonRole.ResetRole,
            ButtonRole.Stretch,
            ButtonRole.ActionRole,
            ButtonRole.ApplyRole | ButtonRole.Reverse,
            ButtonRole.DestructiveRole | ButtonRole.Reverse,
            ButtonRole.AlternateRole | ButtonRole.Reverse,
            ButtonRole.RejectRole | ButtonRole.Reverse,
            ButtonRole.AcceptRole | ButtonRole.Reverse,
            ButtonRole.NoRole | ButtonRole.Reverse,
            ButtonRole.YesRole | ButtonRole.Reverse,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.HelpRole,
            ButtonRole.ResetRole,
            ButtonRole.ApplyRole,
            ButtonRole.ActionRole,
            ButtonRole.Stretch,
            ButtonRole.RejectRole | ButtonRole.Reverse,
            ButtonRole.NoRole | ButtonRole.Reverse,
            ButtonRole.DestructiveRole | ButtonRole.Reverse,
            ButtonRole.AlternateRole | ButtonRole.Reverse,
            ButtonRole.AcceptRole | ButtonRole.Reverse,
            ButtonRole.YesRole | ButtonRole.Reverse,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
    ),
    (
        (
            ButtonRole.ActionRole,
            ButtonRole.YesRole,
            ButtonRole.AcceptRole,
            ButtonRole.AlternateRole,
            ButtonRole.DestructiveRole,
            ButtonRole.NoRole,
            ButtonRole.RejectRole,
            ButtonRole.ApplyRole,
            ButtonRole.ResetRole,
            ButtonRole.HelpRole,
            ButtonRole.Stretch,
            ButtonRole.EOL,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.YesRole,
            ButtonRole.NoRole,
            ButtonRole.AcceptRole,
            ButtonRole.RejectRole,
            ButtonRole.AlternateRole,
            ButtonRole.DestructiveRole,
            ButtonRole.Stretch,
            ButtonRole.ActionRole,
            ButtonRole.ApplyRole,
            ButtonRole.ResetRole,
            ButtonRole.HelpRole,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.AcceptRole,
            ButtonRole.AlternateRole,
            ButtonRole.ApplyRole,
            ButtonRole.ActionRole,
            ButtonRole.YesRole,
            ButtonRole.NoRole,
            ButtonRole.Stretch,
            ButtonRole.ResetRole,
            ButtonRole.DestructiveRole,
            ButtonRole.RejectRole,
            ButtonRole.HelpRole,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.YesRole,
            ButtonRole.NoRole,
            ButtonRole.AcceptRole,
            ButtonRole.RejectRole,
            ButtonRole.AlternateRole,
            ButtonRole.DestructiveRole,
            ButtonRole.ApplyRole,
            ButtonRole.ActionRole,
            ButtonRole.Stretch,
            ButtonRole.ResetRole,
            ButtonRole.HelpRole,
            ButtonRole.EOL,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
        (
            ButtonRole.YesRole,
            ButtonRole.AcceptRole,
            ButtonRole.AlternateRole,
            ButtonRole.DestructiveRole,
            ButtonRole.NoRole,
            ButtonRole.RejectRole,
            ButtonRole.Stretch,
            ButtonRole.ActionRole,
            ButtonRole.ApplyRole,
            ButtonRole.ResetRole,
            ButtonRole.HelpRole,
            ButtonRole.EOL,
            ButtonRole.EOL,
        ),
    ),
)


class QPlatformDialogHelper(QObject):
    accept: Signal = Signal()
    reject: Signal = Signal()

    StyleHint = StyleHint
    DialogCode = DialogCode
    StandardButton = StandardButton
    ButtonRole = ButtonRole
    ButtonLayout = ButtonLayout

    def __new__(cls, *args, **kwargs) -> Self:  # type: ignore[name-defined]
        new_cls = cls
        if cls is QPlatformDialogHelper:
            if sys.platform in ("win32", "cygwin"):
                from utility.gui.qt.adapters.kernel.qplatformdialoghelper.qwindowsdialoghelpers import (
                    QWindowsDialogHelper,
                )

                new_cls = QWindowsDialogHelper
            elif sys.platform == "linux":
                from utility.gui.qt.adapters.kernel.qplatformdialoghelper.qlinuxdialoghelpers import (
                    LinuxFileDialogHelper,
                )

                new_cls = LinuxFileDialogHelper
            elif sys.platform == "darwin":
                from utility.gui.qt.adapters.kernel.qplatformdialoghelper.qmacdialoghelpers import (
                    QMacDialogHelper,
                )

                new_cls = QMacDialogHelper
            else:
                raise NotImplementedError(f"No dialog helper implemented for {sys.platform}")
        return super().__new__(new_cls)

    def exec(self) -> DialogCode:
        raise NotImplementedError()

    def show(self, window_flags: Qt.WindowFlags, window_state: Qt.WindowState, parent: QObject | None):
        raise NotImplementedError()

    def hide(self):
        raise NotImplementedError()

    def selectMimeTypeFilter(self, filter: str) -> None:
        self._selected_mime_type_filter = filter

    def selectedMimeTypeFilter(self) -> str:
        return getattr(self, "_selected_mime_type_filter", "")

    def isSupportedUrl(self, url: QUrl) -> bool:
        return url.isLocalFile()

    def defaultStyleHint(self) -> StyleHint:
        return StyleHint.DialogIsQtWindow

    def styleHint(self) -> StyleHint:
        return self.defaultStyleHint()

    @staticmethod
    def buttonRole(button: StandardButton) -> ButtonRole:
        if button in (
            StandardButton.Ok,
            StandardButton.Save,
            StandardButton.Open,
            StandardButton.SaveAll,
            StandardButton.Retry,
            StandardButton.Ignore,
        ):
            return ButtonRole.AcceptRole
        if button in (StandardButton.Cancel, StandardButton.Close, StandardButton.Abort):
            return ButtonRole.RejectRole
        if button is StandardButton.Discard:
            return ButtonRole.DestructiveRole
        if button is StandardButton.Help:
            return ButtonRole.HelpRole
        if button is StandardButton.Apply:
            return ButtonRole.ApplyRole
        if button in (StandardButton.Yes, StandardButton.YesToAll):
            return ButtonRole.YesRole
        if button in (StandardButton.No, StandardButton.NoToAll):
            return ButtonRole.NoRole
        if button in (StandardButton.RestoreDefaults, StandardButton.Reset):
            return ButtonRole.ResetRole
        return ButtonRole.InvalidRole

    @staticmethod
    def buttonLayout(orientation: Qt.Orientation = Qt.Orientation.Horizontal, policy: ButtonLayout = ButtonLayout.UnknownLayout) -> tuple[ButtonRole, ...]:
        layout_policy = policy
        if layout_policy == ButtonLayout.UnknownLayout:
            if sys.platform == "darwin":
                layout_policy = ButtonLayout.MacLayout
            elif sys.platform.startswith("linux") or os.name == "posix":
                layout_policy = ButtonLayout.KdeLayout
            elif sys.platform.startswith("android"):
                layout_policy = ButtonLayout.AndroidLayout
            else:
                layout_policy = ButtonLayout.WinLayout
        orientation_index = 1 if orientation == Qt.Orientation.Vertical else 0
        return BUTTON_ROLE_LAYOUTS[orientation_index][layout_policy]


class QPlatformColorDialogHelper(QPlatformDialogHelper):
    currentColorChanged: Signal = Signal(QColor)

    @abstractmethod
    def setCurrentColor(self, color: QColor): ...

    @abstractmethod
    def currentColor(self) -> QColor: ...


class QPlatformFileDialogHelper(QPlatformDialogHelper):
    filterRegExp = r"^(.+)\s*\((.*)\)$"

    FileMode = QFileDialog.FileMode
    AcceptMode = QFileDialog.AcceptMode
    Option = QFileDialog.Option

    fileSelected = Signal(QUrl)
    filesSelected = Signal(list)
    currentChanged = Signal(QUrl)
    directoryEntered = Signal(QUrl)
    filterSelected = Signal(str)

    def __init__(self, options: QFileDialogOptions | None = None):
        super().__init__()
        self._options: QFileDialogOptions = options or QFileDialogOptions()
        self._directory: str = ""
        self._selected_files: list[str] = []
        self._selected_name_filter: str = ""

    @staticmethod
    def cleanFilterList(filter: str) -> list[str]:
        if not filter:
            return []

        match = re.match(QPlatformFileDialogHelper.filterRegExp, filter.strip())
        if not match:
            return [filter.strip()]

        pattern_section = match.group(2).strip()
        if not pattern_section:
            return []

        return [pattern for pattern in pattern_section.split() if pattern]

    def options(self) -> QFileDialogOptions:
        return self._options

    def setOptions(self, options: QFileDialogOptions) -> None:
        self._options = options

    def isSupportedUrl(self, url: QUrl) -> bool:
        scheme = url.scheme()
        return scheme in self._options.supportedSchemes() and (url.isLocalFile() or url.isValid())

    def setDirectory(self, directory: str) -> None:
        self._directory = directory
        self._options.setInitialDirectory(QUrl.fromLocalFile(directory))

    def directory(self) -> str:
        return self._directory

    def selectFile(self, filename: QUrl) -> None:
        if filename.isLocalFile():
            self._selected_files = [filename.toLocalFile()]
        else:
            self._selected_files = [filename.toString()]

    def selectedFiles(self) -> list[str]:
        return list(self._selected_files)

    def setFilter(self) -> None:
        pass

    def selectNameFilter(self, filter: str) -> None:  # noqa: A002
        self._selected_name_filter = filter
        self._options.selectNameFilter(filter)

    def selectedNameFilter(self) -> str:
        if self._options.selectedNameFilter():
            return self._options.selectedNameFilter()
        return self._selected_name_filter

    def selectedMimeTypeFilter(self) -> str:
        return self._options.selectedMimeTypeFilter()

    def selectMimeTypeFilter(self, filter: str) -> None:  # noqa: A002
        self._options.selectMimeTypeFilter(filter)

    def defaultNameFilterString(self) -> str:
        return self._options.defaultNameFilterString()


class QPlatformFontDialogHelper(QPlatformDialogHelper):
    currentFontChanged = Signal(QFont)

    @abstractmethod
    def setCurrentFont(self, font: QFont): ...

    @abstractmethod
    def currentFont(self) -> QFont: ...


class QPlatformMessageDialogHelper(QPlatformDialogHelper):
    """Abstract base class for platform-specific message dialog helpers.
    This class defines the interface for creating and managing message dialogs across different platforms.
    """

    StandardButton: QMessageBox.StandardButton = QMessageBox.StandardButton.Ok
    StandardButtons = QMessageBox.StandardButton(0)

    class ButtonRole(Flag):
        """Enum defining the roles of buttons in a message dialog.
        These roles determine the semantic meaning and behavior of buttons.
        """

        InvalidRole = -1  # Invalid button role
        AcceptRole = 0  # Accepting or "OK" role
        RejectRole = 1  # Rejecting or "Cancel" role
        DestructiveRole = 2  # Destructive or dangerous action role
        ActionRole = 3  # Action button role
        HelpRole = 4  # Help button role
        YesRole = 5  # "Yes" button role
        NoRole = 6  # "No" button role
        ResetRole = 7  # Reset to default values role
        ApplyRole = 8  # Apply changes role

    class Icon(Flag):
        """Enum defining the standard icons that can be displayed in a message dialog."""

        NoIcon = 0  # No icon
        Information = 1  # Information icon
        Warning = 2  # Warning icon
        Critical = 3  # Critical or error icon
        Question = 4  # Question icon

    clickedButton = Signal(int)  # Signal emitted when a button is clicked, passing the button's ID

    @abstractmethod
    def setButtons(self, buttons: QMessageBox.StandardButton):
        """Set the standard buttons to be displayed in the dialog.

        :param buttons: A combination of StandardButtons flags
        """
        ...

    @abstractmethod
    def buttons(self) -> QMessageBox.StandardButton:
        """Get the current standard buttons set for the dialog.

        :return: The current StandardButtons flags
        """
        ...

    @abstractmethod
    def setButtonText(self, button: QMessageBox.StandardButton, text: str):
        """Set custom text for a standard button.

        :param button: The StandardButton to modify
        :param text: The new text for the button
        """
        ...

    @abstractmethod
    def buttonText(self, button: QMessageBox.StandardButton) -> str:
        """Get the current text of a standard button.

        :param button: The StandardButton to query
        :return: The current text of the button
        """
        ...

    @abstractmethod
    def setIcon(self, icon: Icon):
        """Set the icon to be displayed in the dialog.

        :param icon: The Icon enum value to set
        """
        ...

    @abstractmethod
    def icon(self) -> Icon:
        """Get the current icon set for the dialog.

        :return: The current Icon enum value
        """
        ...

    @abstractmethod
    def setText(self, text: str):
        """Set the main text of the message dialog.

        :param text: The main message text
        """
        ...

    @abstractmethod
    def text(self) -> str:
        """Get the current main text of the message dialog.

        :return: The current main message text
        """
        ...

    @abstractmethod
    def setInformativeText(self, text: str):
        """Set additional informative text for the dialog.

        :param text: The informative text to set
        """
        ...

    @abstractmethod
    def informativeText(self) -> str:
        """Get the current informative text of the dialog.

        :return: The current informative text
        """
        ...

    @abstractmethod
    def setDetailedText(self, text: str):
        """Set detailed text for the dialog, typically shown in an expandable section.

        :param text: The detailed text to set
        """
        ...

    @abstractmethod
    def detailedText(self) -> str:
        """Get the current detailed text of the dialog.

        :return: The current detailed text
        """
        ...



"""TSLPatcher modification template: OverrideType and PatcherModifications base."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import Literal  # pyright: ignore[reportMissingModuleSource]

    from pykotor.common.misc import Game
    from pykotor.resource.type import SOURCE_TYPES
    from pykotor.tslpatcher.logger import PatchLogger
    from pykotor.tslpatcher.memory import PatcherMemory
    from utility.common.more_collections import CaseInsensitiveDict


class OverrideType:
    """Possible actions for how the patcher should behave when patching a file to a ERF/MOD/RIM while that filename already exists in the Override folder."""

    IGNORE = "ignore"  # Do nothing: don't even check (TSLPatcher default)
    WARN = "warn"  # Log a warning (HoloPatcher default)
    RENAME = "rename"  # Rename the file in the Override folder with the 'old_' prefix. Also logs a warning.


class PatcherModifications(ABC):
    """Abstract base class for TSLPatcher modifications.

    Args:
    ----
        sourcefile (str): The source file for the modifications.
        replace (bool | None, optional): Whether to replace the file. Defaults to None.
        modifiers (list | None, optional): List of modifiers. Defaults to None.

    Attributes:
    ----------
        sourcefile (str): The source file for the modifications.
        sourcefolder (str): The source folder.
        saveas (str): The final name of the file this patch will save as (!SaveAs/!Filename)
        replace_file (bool): Whether to replace the file.
            This bool is True when using syntax Replace#=file_to_replace.ext, and therefore False when File#=file_to_replace.ext syntax is used.
            It is currently unknown whether this takes priority over !ReplaceFile, current PyKotor implementation will prioritize !ReplaceFile
        destination (str): The destination for the patch file.
        action (str): The action for this patch, purely used for logging purposes.
        override_type (str): The override type, see `class OverrideType` above.
        skip_if_not_replace (bool): Determines how !ReplaceFile will be handled.
            TSLPatcher's InstallList and CompileList are the only patchlists that handle replace behavior differently.
            in InstallList/CompileList, if this is True and !ReplaceFile is False or File#=file_to_install.ext, the resource will be skipped if the resource already exists.

    Methods:
    -------
        patch_resource(source, memory, logger, game): Patch the resource defined by the 'source' arg. Returns the bytes data of the result.
        apply(mutable_data, memory, logger, game): Apply this patch's modifications to the mutable_data object argument passed.
        pop_tslpatcher_vars(file_section_dict, default_destination): Parse optional TSLPatcher exclamation point variables.

    Exclamation-point variables:
    ---------------------------
        NOTE: All exclamation-point variables that define a path in TSLPatcher must be backslashed instead of forward-slashed. PyKotor will normalize both slashes though.
        - Top-level variables (e.g. [CompileList] [InstallList] [GFFList])
            !DefaultDestination=relative/path/to/destination/folder - Determines where the destination folder is for top-level patch objects.
                Note: !DefaultDestination is highly undocumented in TSLPatcher so it's unclear whether this matches what TSLPatcher does. I believe it takes priority over InstallList's destinations (excluding !Destination)
        - File-level variables ( e.g. [my_file.nss] )
            !SourceFile=this_file.extension - the name of the file to load. Defaults to 'this_file.ext' when using the `File#=this_file` or `Replace#=this_file` syntax.
            !ReplaceFile=<1 or 0> - Whether to replace the file. Takes priority over `Replace#=this_file.ext` syntax
            !SaveAs=<some_file.tpc> - Determines the final filename of the patch. Defaults to whatever !SourceFile is defined as.
            !Filename=<asdf_file.qwer> - Literally the same as !SaveAs
            !Destination=relative/path/to/destination/folder - The relative path to the folder to save this patched file.
            !OverrideType=<warn or ignore or rename> - How to handle conflict resolution. See `class OverrideType` above.
            !SourceFolder=relative/path/to/tslpatchdata/subfolder - **NEW HOLOPATCHER** support for pathing within the mod's tslpatchdata itself. Currently only used in InstallList.
        NOTE: Some patch lists, albeit rare, have different exclamation-point variables. See tslpatcher/mods/ncs.py and tslpatcher/mods/tlk.py for outliers.
    """  # noqa: E501, W505

    DEFAULT_DESTINATION = "Override"

    @abstractmethod
    def __init__(
        self,
        sourcefile: str,
        replace: bool | None = None,  # noqa: FBT001
        modifiers: list | None = None,
        *,
        destination: str | None = None,
    ):
        self.sourcefile: str = sourcefile
        self.sourcefolder: str = "."
        self.saveas: str = sourcefile
        self.replace_file: bool = bool(replace)
        self.destination: str = self.DEFAULT_DESTINATION if destination is None else destination

        # Full path to source file for copying to tslpatchdata (set by diff engine)
        self._source_filepath: Path | None = None

        self.action: str = "Patch" + " "
        self.override_type: str = OverrideType.WARN
        self.skip_if_not_replace: bool = False  # [InstallList] and [CompileList] only

    @abstractmethod
    def patch_resource(
        self,
        source: SOURCE_TYPES,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ) -> bytes | Literal[True]:
        """If bytes is returned, patch the resource. If True is returned, skip this resource."""

    @abstractmethod
    def apply(
        self,
        mutable_data: Any,
        memory: PatcherMemory,
        logger: PatchLogger,
        game: Game,
    ): ...

    def pop_tslpatcher_vars(
        self,
        file_section_dict: CaseInsensitiveDict[str],
        default_destination: str | None = None,
        default_sourcefolder: str = ".",
    ):
        """All optional TSLPatcher vars that can be parsed for a given patch list."""
        ####
        # NOTE: The second argument passed to the 'pop' function is the default.
        ####

        self.sourcefile = file_section_dict.pop("!SourceFile", self.sourcefile)
        # !SaveAs and !Filename are the same.
        self.saveas = file_section_dict.pop("!Filename", file_section_dict.pop("!SaveAs", self.saveas))

        destination_fallback: str = self.DEFAULT_DESTINATION if default_destination is None else default_destination
        self.destination = file_section_dict.pop("!Destination", destination_fallback)

        # !ReplaceFile=1 is prioritized, see Stoffe's HLFP mod v2.1 for reference.
        replace_file: bool | str = file_section_dict.pop("!ReplaceFile", self.replace_file)
        self.replace_file = convert_to_bool(replace_file)

        # TSLPatcher defaults to "ignore". However realistically, Override file shadowing is
        # a major problem, so HoloPatcher defaults to "warn"
        self.override_type = file_section_dict.pop("!OverrideType", OverrideType.WARN).lower()
        # !SourceFolder: Relative path from mod_path (which is typically the tslpatchdata folder) to source files.
        # Default value "." refers to mod_path itself (the tslpatchdata folder), not its parent.
        # Path resolution: mod_path / sourcefolder / sourcefile
        # For example: if mod_path = "C:/Mod/tslpatchdata" and sourcefolder = ".", then:
        #   - Final path = "C:/Mod/tslpatchdata" / "." / "file.ext" = "C:/Mod/tslpatchdata/file.ext"
        #   - If sourcefolder = "subfolder", then: "C:/Mod/tslpatchdata" / "subfolder" / "file.ext" = "C:/Mod/tslpatchdata/subfolder/file.ext"
        self.sourcefolder = file_section_dict.pop("!SourceFolder", default_sourcefolder)


def convert_to_bool(value: bool | str) -> bool:  # noqa: FBT001
    """Convert a value to boolean.

    The value can be:
    - A boolean (True or False)
    - A string "1" (which should be converted to True)
    - A string "0" (which should be converted to False)

    This function is redundant, but provided for users that may not understand Python.
    """
    return value is True or value == "1"



"""BIF (bif index) package: BIF archive data and binary I/O (read_bif, write_bif, bytes_bif)."""

from __future__ import annotations

from pykotor.resource.formats.bif.bif_data import BIF, BIFResource, BIFType
from pykotor.resource.formats.bif.bif_auto import read_bif, write_bif, bytes_bif
from pykotor.resource.formats.bif.io_bif import BIFBinaryReader, BIFBinaryWriter

__all__ = [
    "BIF",
    "BIFBinaryReader",
    "BIFBinaryWriter",
    "BIFResource",
    "BIFType",
    "bytes_bif",
    "read_bif",
    "write_bif",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import Signal  # pyright: ignore[reportPrivateImportUsage]
from qtpy.QtGui import QFont, QFontDatabase
from qtpy.QtWidgets import QHBoxLayout, QLabel, QSizePolicy, QTextEdit, QVBoxLayout, QWidget

from utility.gui.qt.widgets.font.font_item_widget import FontItemWidget
from utility.gui.qt.widgets.font.size_widget import SizeWidget
from utility.gui.qt.widgets.font.style_widget import StyleWidget


class FontWidget(QWidget):
    sig_font_changed: Signal = Signal(QFont)

    def __init__(
        self,
        font: QFont = QFont("Arial", 10),
    ):
        super().__init__()
        self._current_font: QFont = font
        self._init_ui(font=font)

    def _init_ui(
        self,
        font: QFont,
    ):
        self._preview_text_edit: QTextEdit = QTextEdit()
        self._preview_text_edit.textChanged.connect(self._text_changed)

        self._font_item_widget = FontItemWidget(font)
        self._font_item_widget.sig_item_font_changed.connect(self._font_item_changed_exec)

        self._size_widget = SizeWidget(font)
        self._size_widget.sig_item_size_changed.connect(self._size_item_changed_exec)

        self._style_widget = StyleWidget(font)
        self._style_widget.sig_bold_checked.connect(self._set_bold)
        self._style_widget.sig_italic_checked.connect(self._set_italic)

        self._init_preview_text_edit()

        lay = QHBoxLayout()
        lay.addWidget(self._font_item_widget)
        lay.addWidget(self._size_widget)
        lay.addWidget(self._style_widget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)

        top_widget = QWidget()
        top_widget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(QLabel("Preview"))
        lay.addWidget(self._preview_text_edit)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(5)

        bottom_widget = QWidget()
        bottom_widget.setLayout(lay)
        bottom_widget.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)

        lay = QVBoxLayout()
        lay.addWidget(top_widget)
        lay.addWidget(bottom_widget)
        self.setLayout(lay)

    def _init_preview_text_edit(self):
        font_family: str = self._font_item_widget.get_font_family()
        font_size: str = self._size_widget.get_size()
        bold_f: bool = self._style_widget.is_bold()
        italic_f: bool = self._style_widget.is_italic()
        font: QFont = self._preview_text_edit.currentFont()
        font.setFamily(font_family)
        font.setPointSize(int(font_size))
        font.setBold(bold_f)
        font.setItalic(italic_f)
        self._preview_text_edit.setCurrentFont(font)
        self._preview_text_edit.setText("Sample")

    def _set_bold(
        self,
        f: bool,
    ):
        self._preview_text_edit.selectAll()
        font: QFont = self._preview_text_edit.currentFont()
        font.setBold(f)
        self._preview_text_edit.setCurrentFont(font)

        self._current_font = font
        self.sig_font_changed.emit(self._current_font)

    def _set_italic(
        self,
        f: bool,
    ):
        self._preview_text_edit.selectAll()
        font: QFont = self._preview_text_edit.currentFont()
        font.setItalic(f)
        self._preview_text_edit.setCurrentFont(font)

        self._current_font = font
        self.sig_font_changed.emit(self._current_font)

    def _size_item_changed_exec(
        self,
        size: int,
    ):
        self._preview_text_edit.selectAll()
        font: QFont = self._preview_text_edit.currentFont()
        font.setPointSize(size)
        self._preview_text_edit.setCurrentFont(font)

        self._current_font = font
        self.sig_font_changed.emit(self._current_font)

    def _font_item_changed_exec(
        self,
        font_text: str,
        _fd: object,  # QFontDatabase instance may not be constructible on some Qt6 bindings.
    ):
        self._preview_text_edit.selectAll()
        font: QFont = self._preview_text_edit.currentFont()
        prev_size: int = font.pointSize()
        try:
            styles: list[str] = QFontDatabase.styles(font_text)  # type: ignore[arg-type]
            point_sizes = QFontDatabase.pointSizes  # type: ignore[attr-defined]
        except Exception:  # noqa: BLE001
            fd: QFontDatabase = QFontDatabase()  # pyright: ignore[reportCallIssue]
            styles = fd.styles(font_text)
            point_sizes = fd.pointSizes

        font.setFamily(font_text)

        sizes: list[int] = point_sizes(font_text, styles[0])
        if prev_size in sizes:
            self._size_widget.set_sizes(sizes, prev_size)
            font.setPointSize(prev_size)
        else:
            self._size_widget.set_sizes(sizes, prev_size)
            # font.setPointSize(sizes[0])

        self._preview_text_edit.setCurrentFont(font)
        self._current_font = font
        self.sig_font_changed.emit(self._current_font)

    def get_font(self) -> QFont:
        return self._preview_text_edit.currentFont()

    def set_current_font(
        self,
        font: QFont,
    ):
        self._font_item_widget.set_current_font(font=font)
        self._size_widget.set_current_size(font=font)
        self._style_widget.set_current_style(font=font)

        self._preview_text_edit.setCurrentFont(font)
        self._current_font = font
        self.sig_font_changed.emit(self._current_font)

    def _text_changed(self):
        text: str = self._preview_text_edit.toPlainText()
        if text.strip() != "":
            pass
        else:
            self._preview_text_edit.setCurrentFont(self._current_font)



"""Abstract animation system interfaces.

This module provides abstract base classes for animation that can be implemented
by different rendering backends.

References:
----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers


"""

from __future__ import annotations

from pykotor.engine.animation.base import (
    IAnimationController,
    IAnimationState,
    IAnimationManager,
)

__all__ = [
    "IAnimationController",
    "IAnimationState",
    "IAnimationManager",
]



"""Validation and investigation command implementations for Pykotorcli.

This module provides CLI commands for:
- Checking resource availability (TXI files, 2DA files)
- Validating installations
- Investigating module structure
- Checking missing resources

References:
----------
        Based on swkotor.exe resource validation:
        - CExoResMan::GetResRef - Resource reference lookup
        - CResGFF, CRes2DA, CResTPC - Resource format validators

        scripts/kotor/check_txi_files.py - TXI file checking
        scripts/kotor/check_missing_resources.py - Missing resource checking
        scripts/kotor/investigate_module_structure.py - Module investigation
        Libraries/PyKotor/src/pykotor/tools/validation.py - Core validation functions

"""

from __future__ import annotations

import json
import pathlib

from collections.abc import Sized
from typing import TYPE_CHECKING

from pykotor.cli.console import ok_fail_symbols
from pykotor.common.module import Module
from pykotor.extract.installation import Installation
from pykotor.tools.validation import (
    check_2da_file,
    check_missing_resources_referenced,
    check_txi_files,
    get_installation_summary,
    get_module_referenced_resources,
    investigate_module_structure,
    validate_installation,
)

if TYPE_CHECKING:
    from argparse import Namespace

    from loggerplus import RobustLogger as Logger


def cmd_check_txi(args: Namespace, logger: Logger) -> int:
    """Check if TXI files exist for specific textures.

    Usage:
        pykotorcli check-txi --textures TEXTURE1 TEXTURE2 ...
        pykotorcli check-txi --textures lda_bark04 lda_flr11 --installation "C:/Games/KOTOR"
    """
    try:
        installation = Installation(pathlib.Path(args.installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    if not args.textures:
        logger.error("No textures specified. Use --textures TEXTURE1 TEXTURE2 ...")
        return 1

    logger.info(f"Checking TXI files for {len(args.textures)} textures...")  # noqa: G004
    results = check_txi_files(installation, args.textures)

    ok, fail = ok_fail_symbols()
    found_count = 0
    for tex_name, paths in results.items():
        if paths:
            found_count += 1
            logger.info(f"{ok} {tex_name}.txi: FOUND ({len(paths)} location(s))")  # noqa: G004
            if args.verbose:
                for path in paths[:3]:  # Show first 3 locations
                    logger.info(f"    - {path}")  # noqa: G004
                if len(paths) > 3:
                    logger.info(f"    ... and {len(paths) - 3} more")  # noqa: G004
        else:
            logger.warning(f"{fail} {tex_name}.txi: NOT FOUND")  # noqa: G004

    logger.info(f"\nSummary: {found_count}/{len(args.textures)} TXI files found")  # noqa: G004
    return 0 if found_count == len(args.textures) else 1


def cmd_check_2da(args: Namespace, logger: Logger) -> int:
    """Check if a 2DA file exists in installation.

    Usage:
        pykotorcli check-2da --2da genericdoors --installation "C:/Games/KOTOR"
    """
    try:
        installation = Installation(pathlib.Path(args.two_da_installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    logger.info(f"Checking for 2DA file: {args.two_da_name}")  # noqa: G004
    found, paths = check_2da_file(installation, args.two_da_name)
    ok, fail = ok_fail_symbols()

    if found:
        logger.info(f"{ok} {args.two_da_name}.2da: FOUND")  # noqa: G004
        for path in paths:
            logger.info(f"    - {path}")  # noqa: G004
        return 0
    logger.warning(f"{fail} {args.two_da_name}.2da: NOT FOUND")  # noqa: G004
    return 1


def cmd_validate_installation(args: Namespace, logger: Logger) -> int:
    """Validate a KOTOR installation.

    Usage:
        pykotorcli validate-installation --installation "C:/Games/KOTOR"
    """
    try:
        installation = Installation(pathlib.Path(args.installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    logger.info(f"Validating installation: {installation.path()}")  # noqa: G004
    summary = get_installation_summary(installation, check_essential_files=args.check_essential)
    ok, _fail = ok_fail_symbols()

    if summary["valid"]:
        logger.info(f"{ok} Installation is valid")  # noqa: G004
        if summary.get("game"):
            logger.info(f"  Game: {summary['game']}")  # noqa: G004
        if args.verbose:
            logger.info(f"  Modules: {summary.get('module_count', 0)}, Override files: {summary.get('override_file_count', 0)}")  # noqa: G004
        return 0

    missing_files_raw = summary["missing"]
    assert isinstance(missing_files_raw, (Sized, list)), "missing must be a list or a sized iterable"
    missing_files = missing_files_raw
    if bool(missing_files):
        logger.warning(f"Missing files ({len(missing_files)}):")  # noqa: G004
        for file in missing_files:
            logger.warning(f"  - {file}")  # noqa: G004

    errors_raw = summary["errors"]
    assert isinstance(errors_raw, list), "errors must be a list"
    errors: list[str] = errors_raw  # pyright: ignore[reportGeneralTypeIssues]
    if bool(errors):
        logger.error("Errors:")
        for error in errors:  # type: ignore[union-attr]
            logger.error(f"  - {error}")  # noqa: G004

    return 1


def cmd_investigate_module(args: Namespace, logger: Logger) -> int:
    """Investigate a module's structure.

    Usage:
        pykotorcli investigate-module --module danm13 --installation "C:/Games/KOTOR"
        pykotorcli investigate-module --module danm13 --json output.json
    """
    try:
        installation = Installation(pathlib.Path(args.installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    try:
        module = Module(args.module, installation, use_dot_mod=False)
    except Exception:
        logger.exception(f"Failed to load module '{args.module}'")  # noqa: G004
        return 1

    logger.info(f"Investigating module: {args.module}")  # noqa: G004
    info = investigate_module_structure(module)

    if args.json:
        # Output as JSON
        output_path = pathlib.Path(args.json)
        with output_path.open("w", encoding="utf-8") as f:
            json.dump(info, f, indent=2, default=str)
        logger.info(f"Module information written to {output_path}")  # noqa: G004
        return 0

    # Output as formatted text
    logger.info(f"Module: {info['module_name']}")  # noqa: G004
    logger.info(f"Main RIM: {info['main_rim']}")  # noqa: G004
    logger.info(f"Data RIM: {info['data_rim']}")  # noqa: G004
    logger.info(f"Total resources: {info['total_resources']}")  # noqa: G004
    logger.info("\nResources by type:")
    for res_type, count in sorted(info["resources_by_type"].items(), key=lambda x: x[1], reverse=True):
        logger.info(f"  {res_type}: {count}")  # noqa: G004

    if info["lyt_file"]:
        logger.info(f"\nLYT file: {info['lyt_file']}")  # noqa: G004
        logger.info(f"Rooms: {len(info['rooms'])}")  # noqa: G004
        logger.info(f"Room components: {len(info['room_components'])}")  # noqa: G004

    logger.info(f"\nTextures referenced: {len(info['textures'])}")  # noqa: G004
    logger.info(f"Lightmaps referenced: {len(info['lightmaps'])}")  # noqa: G004

    if args.verbose:
        logger.info(f"\nSample textures ({min(10, len(info['textures']))}):")  # noqa: G004
        for tex in sorted(info["textures"])[:10]:
            logger.info(f"  - {tex}")  # noqa: G004

    return 0


def cmd_check_missing_resources(args: Namespace, logger: Logger) -> int:  # noqa: PLR0912
    """Check if missing resources are referenced by module models.

    Usage:
        pykotorcli check-missing-resources --module danm13 --textures TEX1 TEX2 --lightmaps LM1 LM2
    """
    try:
        installation = Installation(pathlib.Path(args.installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    try:
        module = Module(args.module, installation, use_dot_mod=False)
    except Exception:
        logger.exception(f"Failed to load module '{args.module}'")  # noqa: G004
        return 1

    textures_list = args.textures or []
    lightmaps_list = args.lightmaps or []

    if not textures_list and not lightmaps_list:
        logger.error("No resources specified. Use --textures and/or --lightmaps")
        return 1

    logger.info(f"Checking missing resources in module: {args.module}")  # noqa: G004
    results = check_missing_resources_referenced(
        module,
        missing_textures=textures_list,
        missing_lightmaps=lightmaps_list,
    )

    ok, fail = ok_fail_symbols()
    referenced_count = 0
    if textures_list:
        logger.info(f"\nTextures ({len(textures_list)}):")  # noqa: G004
        for tex in textures_list:
            is_referenced = results.get(tex, False)
            if is_referenced:
                referenced_count += 1
                logger.info(f"  {ok} {tex}: REFERENCED")  # noqa: G004
            else:
                logger.warning(f"  {fail} {tex}: NOT REFERENCED")  # noqa: G004

    if lightmaps_list:
        logger.info(f"\nLightmaps ({len(lightmaps_list)}):")  # noqa: G004
        for lm in lightmaps_list:
            is_referenced = results.get(lm, False)
            if is_referenced:
                referenced_count += 1
                logger.info(f"  {ok} {lm}: REFERENCED")  # noqa: G004
            else:
                logger.warning(f"  {fail} {lm}: NOT REFERENCED")  # noqa: G004

    total = len(textures_list) + len(lightmaps_list)
    logger.info(f"\nSummary: {referenced_count}/{total} resources are referenced")  # noqa: G004
    return 0 if referenced_count == total else 1


def cmd_module_resources(args: Namespace, logger: Logger) -> int:
    """Get all resources referenced by a module's models.

    Usage:
        pykotorcli module-resources --module danm13 --installation "C:/Games/KOTOR"
    """
    try:
        installation = Installation(pathlib.Path(args.installation))
    except Exception:
        logger.exception("Invalid installation path")
        return 1

    try:
        module = Module(args.module, installation, use_dot_mod=False)
    except Exception:
        logger.exception(f"Failed to load module '{args.module}'")  # noqa: G004
        return 1

    logger.info(f"Scanning models in module: {args.module}")  # noqa: G004
    textures, lightmaps = get_module_referenced_resources(module)

    logger.info(f"\nTotal textures referenced: {len(textures)}")  # noqa: G004
    logger.info(f"Total lightmaps referenced: {len(lightmaps)}")  # noqa: G004

    if args.output:
        output_path = pathlib.Path(args.output)
        data = {
            "module": args.module,
            "textures": sorted(textures),
            "lightmaps": sorted(lightmaps),
        }
        with output_path.open("w", encoding="utf-8") as f:
            json.dump(data, f, indent=2)
        logger.info(f"Resource list written to {output_path}")  # noqa: G004
    elif args.verbose:
        logger.info(f"\nTextures ({min(20, len(textures))}):")  # noqa: G004
        for tex in sorted(textures)[:20]:
            logger.info(f"  - {tex}")  # noqa: G004
        if len(textures) > 20:
            logger.info(f"  ... and {len(textures) - 20} more")  # noqa: G004

        logger.info(f"\nLightmaps ({min(20, len(lightmaps))}):")  # noqa: G004
        for lm in sorted(lightmaps)[:20]:
            logger.info(f"  - {lm}")  # noqa: G004
        if len(lightmaps) > 20:
            logger.info(f"  ... and {len(lightmaps) - 20} more")  # noqa: G004

    return 0



"""Unified read/write and type dispatch for KotOR resources.

This module provides read_resource, resource_to_bytes, and dismantle_generic so callers
can work with sources and high-level types (ARE, DLG, GIT, etc.) without format-specific imports.
"""
from __future__ import annotations

import os

from contextlib import suppress
from pathlib import PurePath
from typing import TYPE_CHECKING, Union

from pykotor.extract.file import ResourceIdentifier
from pykotor.resource.formats.bwm import BWM, bytes_bwm, read_bwm
from pykotor.resource.formats.erf import ERF, bytes_erf, read_erf
from pykotor.resource.formats.gff import GFF, GFFContent, bytes_gff, read_gff
from pykotor.resource.formats.lip import LIP, bytes_lip, read_lip
from pykotor.resource.formats.ltr import LTR, bytes_ltr, read_ltr
from pykotor.resource.formats.lyt import LYT, bytes_lyt, read_lyt
from pykotor.resource.formats.mdl import MDL, bytes_mdl, read_mdl
from pykotor.resource.formats.ncs import NCS, bytes_ncs, read_ncs
from pykotor.resource.formats.rim import RIM, bytes_rim, read_rim
from pykotor.resource.formats.ssf import SSF, bytes_ssf, read_ssf
from pykotor.resource.formats.tlk import TLK, bytes_tlk, read_tlk
from pykotor.resource.formats.tpc import TPC, bytes_tpc, read_tpc
from pykotor.resource.formats.twoda import TwoDA, bytes_2da, read_2da
from pykotor.resource.formats.vis import VIS, bytes_vis, read_vis
from pykotor.resource.generics.are import ARE, dismantle_are
from pykotor.resource.generics.dlg import DLG, dismantle_dlg
from pykotor.resource.generics.git import GIT, dismantle_git
from pykotor.resource.generics.ifo import IFO, dismantle_ifo
from pykotor.resource.generics.jrl import JRL, dismantle_jrl
from pykotor.resource.generics.pth import PTH, dismantle_pth
from pykotor.resource.generics.utc import UTC, dismantle_utc
from pykotor.resource.generics.utd import UTD, dismantle_utd
from pykotor.resource.generics.ute import UTE, dismantle_ute
from pykotor.resource.generics.utm import UTM, dismantle_utm
from pykotor.resource.generics.utp import UTP, dismantle_utp
from pykotor.resource.generics.uts import UTS, dismantle_uts
from pykotor.resource.generics.utt import UTT, dismantle_utt
from pykotor.resource.generics.utw import UTW, dismantle_utw
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pykotor.resource.type import SOURCE_TYPES


def _ncs_to_bytes(ncs: NCS) -> bytes:
    return bytes(bytes_ncs(ncs))


def _read_ncs_as_bytes(source: SOURCE_TYPES) -> bytes:
    return _ncs_to_bytes(read_ncs(source))


def _try_read_with(readers: tuple[callable, ...], source: SOURCE_TYPES) -> bytes | None:
    for reader in readers:
        with suppress(OSError, ValueError):
            return reader(source)
    return None


_UNKNOWN_RESOURCE_READERS: tuple[callable, ...] = (
    lambda source: bytes_tlk(read_tlk(source)),
    lambda source: bytes_ssf(read_ssf(source)),
    lambda source: bytes_2da(read_2da(source)),
    lambda source: bytes_lip(read_lip(source)),
    lambda source: bytes_tpc(read_tpc(source)),
    lambda source: bytes_erf(read_erf(source)),
    lambda source: bytes_rim(read_rim(source)),
    _read_ncs_as_bytes,
    lambda source: bytes_gff(read_gff(source)),
    lambda source: bytes_mdl(read_mdl(source)),
    lambda source: bytes_vis(read_vis(source)),
    lambda source: bytes_lyt(read_lyt(source)),
    lambda source: bytes_ltr(read_ltr(source)),
    lambda source: bytes_bwm(read_bwm(source)),
)


_GENERIC_DISMANTLERS: tuple[tuple[type, callable], ...] = (
    (ARE, dismantle_are),
    (DLG, dismantle_dlg),
    (GIT, dismantle_git),
    (IFO, dismantle_ifo),
    (JRL, dismantle_jrl),
    (PTH, dismantle_pth),
    (UTC, dismantle_utc),
    (UTD, dismantle_utd),
    (UTE, dismantle_ute),
    (UTM, dismantle_utm),
    (UTP, dismantle_utp),
    (UTS, dismantle_uts),
    (UTT, dismantle_utt),
    (UTW, dismantle_utw),
)


_RESOURCE_SERIALIZERS: tuple[tuple[type, callable], ...] = (
    (BWM, bytes_bwm),
    (GFF, bytes_gff),
    (ERF, bytes_erf),
    (LIP, bytes_lip),
    (LTR, bytes_ltr),
    (LYT, bytes_lyt),
    (MDL, bytes_mdl),
    (NCS, _ncs_to_bytes),
    (RIM, bytes_rim),
    (SSF, bytes_ssf),
    (TLK, bytes_tlk),
    (TPC, bytes_tpc),
    (TwoDA, bytes_2da),
    (VIS, bytes_vis),
)


def _read_ext_ssf(source: SOURCE_TYPES) -> bytes:
    return bytes_ssf(read_ssf(source))


def _read_ext_2da(source: SOURCE_TYPES) -> bytes:
    return bytes_2da(read_2da(source))


def _read_ext_lip(source: SOURCE_TYPES) -> bytes:
    return bytes_lip(read_lip(source))


def _read_ext_rim(source: SOURCE_TYPES) -> bytes:
    return bytes_rim(read_rim(source))


def _read_ext_ncs(source: SOURCE_TYPES) -> bytes:
    return bytes(bytes_ncs(read_ncs(source)))


def _read_ext_mdl(source: SOURCE_TYPES) -> bytes:
    return bytes_mdl(read_mdl(source))


def _read_ext_vis(source: SOURCE_TYPES) -> bytes:
    return bytes_vis(read_vis(source))


def _read_ext_lyt(source: SOURCE_TYPES) -> bytes:
    return bytes_lyt(read_lyt(source))


def _read_ext_ltr(source: SOURCE_TYPES) -> bytes:
    return bytes_ltr(read_ltr(source))


_READ_RESOURCE_BY_EXT: dict[str, callable] = {
    "ssf": _read_ext_ssf,
    "2da": _read_ext_2da,
    "lip": _read_ext_lip,
    "rim": _read_ext_rim,
    "ncs": _read_ext_ncs,
    "mdl": _read_ext_mdl,
    "vis": _read_ext_vis,
    "lyt": _read_ext_lyt,
    "ltr": _read_ext_ltr,
}


def read_resource(  # noqa: C901, PLR0911, PLR0912
    source: SOURCE_TYPES,
    resource_type: ResourceType | None = None,
) -> bytes:
    """Reads a resource from a source and returns it as bytes.

    This is a convenience method to make getting the resource's data easier.
    Can handle various formats (XML/CSV/JSON etc).

    Args:
    ----
        source: SOURCE_TYPES: The source of the resource
        resource_type: ResourceType | None: The type of the resource

    Returns:
    -------
        bytes: The resource data as bytes
    """
    source_path: os.PathLike | str | None = None
    with suppress(Exception):
        if isinstance(source, (os.PathLike, str)):
            source_path = source
            if not resource_type:
                _resname, resource_type = ResourceIdentifier.from_path(source).unpack()

    if not resource_type:
        return read_unknown_resource(source)

    try:  # dispatch by category/extension for maintainability
        ext_obj = PurePath(resource_type.extension)
        resource_ext = ext_obj.stem.lower() if "." in ext_obj.name else ext_obj.suffix.lower()[1:]

        # Category-based dispatch
        if resource_type.category == "Talk Tables":
            return bytes_tlk(read_tlk(source))
        if resource_type.category == "Walkmeshes":
            return bytes_bwm(read_bwm(source))
        if resource_type in {ResourceType.TGA, ResourceType.TPC}:
            return bytes_tpc(read_tpc(source))

        # Extension-based dispatch (O(1) lookup, single place to add new types)
        handler = _READ_RESOURCE_BY_EXT.get(resource_ext)
        if handler is not None:
            return handler(source)

        if ResourceType.from_extension(resource_ext) in (ResourceType.ERF, ResourceType.MOD, ResourceType.SAV):
            return bytes_erf(read_erf(source))
        if resource_type.extension.upper() in GFFContent.__members__:
            return bytes_gff(read_gff(source))
    except Exception as e:  # pylint: disable=W0718  # noqa: BLE001
        new_err = ValueError(f"Could not load resource '{source_path}' as resource type '{resource_type}")
        print((new_err.__class__.__name__, str(new_err)))
        raise new_err from e

    msg = f"Resource type {resource_type!r} is not supported by this library."
    raise ValueError(msg)


def read_unknown_resource(  # noqa: PLR0911
    source: SOURCE_TYPES,
) -> bytes:
    read_result = _try_read_with(_UNKNOWN_RESOURCE_READERS, source)
    if read_result is not None:
        return read_result
    msg = "Source resource data not recognized as any kotor file formats."
    raise ValueError(msg)


GFF_GENERICS = Union[ARE, DLG, GIT, IFO, JRL, PTH, UTC, UTD, UTE, UTM, UTP, UTS, UTW]


def dismantle_generic(  # noqa: PLR0911, C901, PLR0912, ANN201
    generic: GFF_GENERICS,
) -> GFF:
    """Returns a GFF object from a constructed obj.

    Args:
    ----
        generic: The constructed loaded object to deconstruct into GFF (e.g. DLG, PTH, IFO...)

    Returns:
    -------
        A deconstructed_gff
    """
    for generic_type, dismantler in _GENERIC_DISMANTLERS:
        if isinstance(generic, generic_type):
            return dismantler(generic)
    if isinstance(generic, GFF):
        return generic  # Guess whoever called this didn't get the memo.
    raise TypeError(f"Could not dismantle generic, invalid object passed ({generic}) of type '{type(generic).__name__}' was unexpected.")


def resource_to_bytes(  # noqa: PLR0912, C901, PLR0911
    resource: BWM | ERF | GFF | LIP | LTR | LYT | MDL | NCS | RIM | SSF | TLK | TPC | TwoDA | VIS | GFF_GENERICS,
) -> bytes:
    if isinstance(resource, GFF_GENERICS):
        return bytes_gff(dismantle_generic(resource))
    for resource_type, serializer in _RESOURCE_SERIALIZERS:
        if isinstance(resource, resource_type):
            return serializer(resource)
    raise TypeError(f"Invalid resource {resource} of type '{type(resource).__name__}' passed to `resource_to_bytes`.")



"""Scene graph node: transform hierarchy and optional mesh for OpenGL rendering."""

from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING

from pykotor.gl.glm_compat import Vector3, Vector4, decompose, mat4, mat4_cast, quat, translate

if TYPE_CHECKING:
    from pykotor.gl.models.mesh import Mesh
    from pykotor.gl.scene import Scene
    from pykotor.gl.shader import Shader


class Node:
    def __init__(
        self,
        scene: Scene,
        parent: Node | None,
        name: str,
    ):
        self._scene: Scene = scene
        self._parent: Node | None = parent
        self.name: str = name
        self._transform: mat4 = mat4()
        self._position: Vector3 = Vector3()
        self._rotation: quat = quat()
        self.children: list[Node] = []
        self.render: bool = True
        self.mesh: Mesh | None = None

        self._recalc_transform()

    def root(self) -> Node | None:
        ancestor: Node | None = self._parent
        while ancestor:
            ancestor = ancestor._parent  # noqa: SLF001
        return ancestor

    def ancestors(self) -> list[Node]:
        ancestors: list[Node] = []
        ancestor: Node | None = self._parent
        while ancestor:
            ancestors.append(ancestor)
            ancestor = ancestor._parent  # noqa: SLF001
        return list(reversed(ancestors))

    def global_position(self) -> Vector3:
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * translate(ancestor._position)  # noqa: SLF001
            transform = transform * mat4_cast(ancestor._rotation)  # noqa: SLF001
        position = Vector3()
        decompose(transform, Vector3(), quat(), position, Vector3(), Vector4())  # pyright: ignore[reportCallIssue, reportArgumentType]
        return position

    def global_rotation(self) -> quat:
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * translate(ancestor._position)  # noqa: SLF001
            transform = transform * mat4_cast(ancestor._rotation)  # noqa: SLF001
        rotation = quat()
        decompose(transform, Vector3(), rotation, Vector3(), Vector3(), Vector4())  # pyright: ignore[reportCallIssue, reportArgumentType]
        return rotation

    def global_transform(self) -> mat4:
        ancestors: list[Node] = [*self.ancestors(), self]
        transform = mat4()
        for ancestor in ancestors:
            transform = transform * translate(ancestor._position)  # noqa: SLF001
            transform = transform * mat4_cast(ancestor._rotation)  # noqa: SLF001
        return transform

    def transform(self) -> mat4:
        return copy(self._transform)

    def _recalc_transform(self):
        self._transform = translate(self._position) * mat4_cast(quat(self._rotation))

    def position(self) -> Vector3:
        return copy(self._position)

    def set_position(self, x: float, y: float, z: float):
        self._position = Vector3(x, y, z)
        self._recalc_transform()

    def rotation(self) -> quat:
        return copy(self._rotation)

    def set_rotation(
        self,
        pitch: float,
        yaw: float,
        roll: float,
    ):
        self._rotation = quat(Vector3(pitch, yaw, roll))
        self._recalc_transform()

    def draw(
        self,
        shader: Shader,
        transform: mat4,
        override_texture: str | None = None,
    ):
        transform = transform * self._transform

        if self.mesh and self.render:
            self.mesh.draw(shader, transform, override_texture)

        for child in self.children:
            child.draw(shader, transform, override_texture=override_texture)



"""VIS (visibility) ASCII read/write: room-to-room visibility for occlusion culling."""

from __future__ import annotations

import os

from typing import TYPE_CHECKING

from pykotor.resource.formats.vis.vis_data import VIS
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose

if TYPE_CHECKING:
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


class VISAsciiReader(ResourceReader):
    """Reads VIS (Visibility) files.

    VIS files define which rooms are visible from other rooms, used for occlusion culling
    and level-of-detail management in KotOR modules.

    References:
    ----------
        Based on swkotor.exe VIS structure:
        - LoadVisibility @ 0x004568d0 - Loads VIS file for visibility culling
          * Parses ASCII format
          * Reads parent room names and child room counts
          * Builds visibility graph for occlusion culling
        - "%s/%s.VIS" file path format - Used by engine to locate VIS files
        - ".vis" file extension - VIS file extension
        - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
        Note: VIS files define which rooms are visible from other rooms, used for occlusion culling
        and level-of-detail management in KotOR modules. VIS files are ASCII text files with the
        following format: parent room names followed by indented child room names.


    """

    def __init__(self, source: SOURCE_TYPES, offset: int = 0, size: int = 0):
        super().__init__(source, offset, size)
        self._vis: VIS | None = None
        self._lines: list[str] = []

    @autoclose
    def load(self, *, auto_close: bool = True) -> VIS:  # noqa: FBT001, FBT002, ARG002
        self._vis = VIS()
        self._lines = self._reader.read_string(self._reader.size()).splitlines()

        pairs = []

        iterator = iter(self._lines)
        for line in iterator:
            tokens: list[str] = line.split()

            # Skip empty lines
            if not tokens:
                continue

            # Use a named constant for the magic value 2
            VERSION_HEADER_TOKEN_INDEX = 1  # Index in VIS ASCII lines where a version string may appear
            # Check if this is a version header line (e.g., "room V3.28")
            if len(tokens) >= VERSION_HEADER_TOKEN_INDEX + 1 and tokens[VERSION_HEADER_TOKEN_INDEX].startswith("V"):
                # This is a version header, skip it
                # Format appears to be: roomname Version
                continue

            when_inside: str = tokens[0]
            self._vis.add_room(when_inside)

            # Try to parse the count, provide better error if it fails
            try:
                count = int(tokens[1])
            except (ValueError, IndexError) as e:
                msg = f"Invalid VIS format: expected room count, got '{tokens[1] if len(tokens) > 1 else '(missing)'}' for room '{when_inside}'"
                raise ValueError(msg) from e

            for _ in range(count):
                show = next(iterator).split()[0]
                pairs.append((when_inside, show))

        for when_inside, show in pairs:
            if when_inside not in self._vis.all_rooms():
                self._vis.add_room(when_inside)
            if show not in self._vis.all_rooms():
                self._vis.add_room(show)
            self._vis.set_visible(when_inside, show, visible=True)

        return self._vis


class VISAsciiWriter(ResourceWriter):
    def __init__(self, vis: VIS, target: TARGET_TYPES):
        super().__init__(target)
        self._vis: VIS = vis

    @autoclose
    def write(self, *, auto_close: bool = True):  # noqa: FBT001, FBT002, ARG002  # pyright: ignore[reportUnusedParameters]
        for observer, observed in self._vis:
            self._writer.write_string(f"{observer} {len(observed)}{os.linesep}")
            for room in observed:
                self._writer.write_string(f"  {room}{os.linesep}")



"""QExtendedInformation Python adapter matching Qt6 C++ source exactly.

This class matches qfileinfogatherer_p.h lines 42-122 (QExtendedInformation class).

C++ class definition:
class QExtendedInformation {
public:
    enum Type { Dir, File, System };
    QExtendedInformation() {}
    QExtendedInformation(const QFileInfo &info) : mFileInfo(info) {}
    // ... member functions ...
private:
    QFileInfo mFileInfo;
    QString displayType;
    QIcon icon;
};
"""

from __future__ import annotations

import os

from typing import TYPE_CHECKING

from qtpy.QtCore import QFileInfo, QTimeZone  # noqa: E402
from qtpy.QtGui import QIcon  # noqa: E402

if TYPE_CHECKING:
    from qtpy.QtCore import QDateTime, QFileDevice


class PyQExtendedInformation:
    """Python adapter for QExtendedInformation matching Qt6 C++ source exactly.

    Matches qfileinfogatherer_p.h lines 42-122.
    """

    # Matches C++ line 44: enum Type { Dir, File, System };
    class Type:
        """Type enum matching C++ QExtendedInformation::Type exactly."""

        Dir = 0
        File = 1
        System = 2

    # Convenience constants matching C++ enum values
    Dir: int = Type.Dir
    File: int = Type.File
    System: int = Type.System

    def __init__(self, info: QFileInfo | None = None):
        """Initialize QExtendedInformation matching C++ constructors exactly.

        Matches C++ lines 46-47:
        QExtendedInformation() {}
        QExtendedInformation(const QFileInfo &info) : mFileInfo(info) {}
        """
        # Matches C++ line 121: QFileInfo mFileInfo;
        if info is None:
            self.mFileInfo: QFileInfo = QFileInfo()
        else:
            self.mFileInfo = QFileInfo(info)
        # Matches C++ line 117: QString displayType;
        self.displayType: str = ""
        # Matches C++ line 118: QIcon icon;
        self.icon: QIcon = QIcon()

    def isDir(self) -> bool:
        """Return true if type() == Dir.

        Matches C++: inline bool isDir() { return type() == Dir; }
        """
        return self.type() == PyQExtendedInformation.Dir

    def isFile(self) -> bool:
        """Return true if type() == File.

        Matches C++: inline bool isFile() { return type() == File; }
        """
        return self.type() == PyQExtendedInformation.File

    def isSystem(self) -> bool:
        """Return true if type() == System.

        Matches C++: inline bool isSystem() { return type() == System; }
        """
        return self.type() == PyQExtendedInformation.System

    def __eq__(self, other: object) -> bool:
        """Equality operator matching C++ exactly.

        Matches C++ lines 53-58:
        bool operator ==(const QExtendedInformation &fileInfo) const {
           return mFileInfo == fileInfo.mFileInfo
           && displayType == fileInfo.displayType
           && permissions() == fileInfo.permissions()
           && lastModified(QTimeZone::UTC) == fileInfo.lastModified(QTimeZone::UTC);
        }
        """
        if not isinstance(other, PyQExtendedInformation):
            return False
        return (
            self.mFileInfo == other.mFileInfo
            and self.displayType == other.displayType
            and self.permissions() == other.permissions()
            and self.lastModified(QTimeZone.UTC) == other.lastModified(QTimeZone.UTC)
        )

    def isCaseSensitive(self) -> bool:
        """Return case sensitivity.

        Matches C++ lines 60-65 (QT_NO_FSFILEENGINE version not implemented,
        using platform check instead).
        """
        # C++ uses qt_isCaseSensitive from QFileInfoPrivate, we use platform check
        return os.name == "posix"

    def permissions(self) -> QFileDevice.Permissions:
        """Return file permissions.

        Matches C++ lines 67-69:
        QFile::Permissions permissions() const {
            return mFileInfo.permissions();
        }
        """
        return self.mFileInfo.permissions()

    def type(self) -> int:
        """Return Type enum value.

        Matches C++ lines 71-82:
        Type type() const {
            if (mFileInfo.isDir()) {
                return QExtendedInformation::Dir;
            }
            if (mFileInfo.isFile()) {
                return QExtendedInformation::File;
            }
            if (!mFileInfo.exists() && mFileInfo.isSymLink()) {
                return QExtendedInformation::System;
            }
            return QExtendedInformation::System;
        }
        """
        if self.mFileInfo.isDir():
            return PyQExtendedInformation.Dir
        if self.mFileInfo.isFile():
            return PyQExtendedInformation.File
        if not self.mFileInfo.exists() and self.mFileInfo.isSymLink():
            return PyQExtendedInformation.System
        return PyQExtendedInformation.System

    def isSymLink(self, ignoreNtfsSymLinks: bool = False) -> bool:  # noqa: FBT001, FBT002
        """Check if symlink, matching C++ exactly.

        Matches C++ lines 84-92:
        bool isSymLink(bool ignoreNtfsSymLinks = false) const
        {
            if (ignoreNtfsSymLinks) {
        #ifdef Q_OS_WIN
                return !mFileInfo.suffix().compare(QLatin1StringView("lnk"), Qt::CaseInsensitive);
        #endif
            }
            return mFileInfo.isSymLink();
        }
        """
        if ignoreNtfsSymLinks:
            if os.name == "nt":
                # C++: !mFileInfo.suffix().compare(QLatin1StringView("lnk"), Qt::CaseInsensitive)
                # This means: suffix equals "lnk" (case insensitive)
                return self.mFileInfo.suffix().lower() == "lnk"
        return self.mFileInfo.isSymLink()

    def isHidden(self) -> bool:
        """Return true if file is hidden.

        Matches C++ lines 94-96:
        bool isHidden() const {
            return mFileInfo.isHidden();
        }
        """
        return self.mFileInfo.isHidden()

    def fileInfo(self) -> QFileInfo:
        """Return QFileInfo.

        Matches C++ lines 98-100:
        QFileInfo fileInfo() const {
            return mFileInfo;
        }
        """
        return self.mFileInfo

    def lastModified(self, tz: QTimeZone) -> QDateTime:
        """Return last modified time in timezone.

        Matches C++ lines 102-104:
        QDateTime lastModified(const QTimeZone &tz) const {
            return mFileInfo.lastModified(tz);
        }
        """
        return self.mFileInfo.lastModified(tz)

    def size(self) -> int:
        """Return file size.

        Matches C++ lines 106-115:
        qint64 size() const {
            qint64 size = -1;
            if (type() == QExtendedInformation::Dir)
                size = 0;
            if (type() == QExtendedInformation::File)
                size = mFileInfo.size();
            if (!mFileInfo.exists() && !mFileInfo.isSymLink())
                size = -1;
            return size;
        }
        """
        size = -1
        if self.type() == PyQExtendedInformation.Dir:
            size = 0
        if self.type() == PyQExtendedInformation.File:
            size = self.mFileInfo.size()
        if not self.mFileInfo.exists() and not self.mFileInfo.isSymLink():
            size = -1
        return size



"""Panda3D implementation of animation system.

This module implements the abstract animation interfaces using Panda3D-specific types.

References:
----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Libraries/PyKotor/src/pykotor/engine/animation/base.py - Abstract interfaces
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers


"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from panda3d.core import Quat, Vec3

from pykotor.engine.animation.base import (
    IAnimationController,
    IAnimationManager,
    IAnimationState,
)

if TYPE_CHECKING:
    from panda3d.core import NodePath

    from pykotor.resource.formats.mdl.mdl_data import (
        MDL,
        MDLAnimation,
        MDLController,
        MDLControllerRow,
    )


class Panda3DAnimationController(IAnimationController):
    """Base Panda3D animation controller.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Libraries/PyKotor/src/pykotor/engine/animation/base.py:23-53 - IAnimationController
        /panda3d/panda3d-docs/programming/scene-graph/common-state-changes - NodePath transforms
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/OdysseyController.ts:18-47


    """

    def __init__(self, mdl_controller: MDLController):
        """Initialize the controller.

        Args:
        ----
            mdl_controller: MDL controller data from PyKotor
        """
        self.controller_type = mdl_controller.controller_type
        self.is_bezier = mdl_controller.is_bezier
        self.keyframes: list[tuple[float, Any]] = []

        # Convert MDL controller rows to keyframes
        for row in mdl_controller.rows:
            value = self._convert_row_data(row)
            self.keyframes.append((row.time, value))

    def _convert_row_data(self, row: MDLControllerRow) -> Any:
        """Convert MDL row to controller-specific format."""
        raise NotImplementedError("Subclasses must implement _convert_row_data")

    def get_value_at_time(self, time: float) -> Any:
        """Get interpolated value at time."""
        if not self.keyframes:
            return None

        if time <= self.keyframes[0][0]:
            return self.keyframes[0][1]
        if time >= self.keyframes[-1][0]:
            return self.keyframes[-1][1]

        # Find surrounding keyframes
        last_idx = 0
        for i, (kf_time, _) in enumerate(self.keyframes):
            if kf_time > time:
                break
            last_idx = i

        next_idx = min(last_idx + 1, len(self.keyframes) - 1)

        last_time, last_value = self.keyframes[last_idx]
        next_time, next_value = self.keyframes[next_idx]

        if last_idx == next_idx:
            return last_value

        duration = next_time - last_time
        if duration == 0:
            return last_value

        factor = (time - last_time) / duration

        return self._interpolate(last_value, next_value, factor)

    def _interpolate(self, last_value: Any, next_value: Any, factor: float) -> Any:
        """Interpolate between values."""
        raise NotImplementedError("Subclasses must implement _interpolate")


class Panda3DPositionController(Panda3DAnimationController):
    """Position animation controller using Panda3D Vec3.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - NodePath.setPos()
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/PositionController.ts:17-122


    """

    def _convert_row_data(self, row: MDLControllerRow) -> Vec3:
        """Convert to Panda3D Vec3."""
        if len(row.data) >= 3:
            return Vec3(row.data[0], row.data[1], row.data[2])
        return Vec3(0, 0, 0)

    def apply(self, node: NodePath, value: Vec3) -> None:
        """Apply position to node.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - myNodePath.setPos(X, Y, Z)


        """
        node.setPos(value)

    def _interpolate(self, last_value: Vec3, next_value: Vec3, factor: float) -> Vec3:
        """Linear interpolation of Vec3.

        References:
        ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/PositionController.ts:106-117


        """
        return Vec3(
            last_value.x + (next_value.x - last_value.x) * factor,
            last_value.y + (next_value.y - last_value.y) * factor,
            last_value.z + (next_value.z - last_value.z) * factor,
        )


class Panda3DOrientationController(Panda3DAnimationController):
    """Orientation animation controller using Panda3D Quat.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - Quaternion and SLERP
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/OrientationController.ts:17-97


    """

    def _convert_row_data(self, row: MDLControllerRow) -> Quat:
        """Convert to Panda3D Quat.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs/programming/physics/ode/worlds-bodies-masses - setQuaternion()



        Notes:
        -----
            MDL stores quaternions as (x, y, z, w)
            Panda3D Quat constructor is (w, x, y, z)
        """
        if len(row.data) >= 4:
            # Reorder: MDL (x,y,z,w) -> Panda3D Quat(w,x,y,z)
            return Quat(row.data[3], row.data[0], row.data[1], row.data[2])
        return Quat(1, 0, 0, 0)  # Identity

    def apply(self, node: NodePath, value: Quat) -> None:
        """Apply orientation to node.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - LerpQuatInterval accepts Quat



        Notes:
        -----
            We convert Quat to HPR for setHpr() since that's the standard API.
        """
        # Convert quaternion to HPR (heading, pitch, roll)
        hpr = value.getHpr()
        node.setHpr(hpr)

    def _interpolate(self, last_value: Quat, next_value: Quat, factor: float) -> Quat:
        """Spherical linear interpolation (SLERP).

        References:
        ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/OrientationController.ts:86-87



        Notes:
        -----
            Panda3D doesn't have a built-in SLERP method on Quat,
            so we implement it manually.
        """
        # Compute dot product
        dot = last_value.getR() * next_value.getR() + last_value.getI() * next_value.getI() + last_value.getJ() * next_value.getJ() + last_value.getK() * next_value.getK()

        # If dot product is negative, negate one quaternion to take shorter path
        if dot < 0:
            next_value = Quat(-next_value.getR(), -next_value.getI(), -next_value.getJ(), -next_value.getK())
            dot = -dot

        # Clamp dot product
        if dot > 0.9995:
            # Quaternions are very close, use linear interpolation
            result = Quat(
                last_value.getR() + factor * (next_value.getR() - last_value.getR()),
                last_value.getI() + factor * (next_value.getI() - last_value.getI()),
                last_value.getJ() + factor * (next_value.getJ() - last_value.getJ()),
                last_value.getK() + factor * (next_value.getK() - last_value.getK()),
            )
            result.normalize()
            return result

        # Calculate angle
        import math

        theta_0 = math.acos(dot)
        theta = theta_0 * factor

        sin_theta = math.sin(theta)
        sin_theta_0 = math.sin(theta_0)

        s0 = math.cos(theta) - dot * sin_theta / sin_theta_0
        s1 = sin_theta / sin_theta_0

        return Quat(
            s0 * last_value.getR() + s1 * next_value.getR(),
            s0 * last_value.getI() + s1 * next_value.getI(),
            s0 * last_value.getJ() + s1 * next_value.getJ(),
            s0 * last_value.getK() + s1 * next_value.getK(),
        )


class Panda3DScaleController(Panda3DAnimationController):
    """Scale animation controller.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - NodePath.setScale()
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/ScaleController.ts:17-120


    """

    def _convert_row_data(self, row: MDLControllerRow) -> float:
        """Convert to scale float."""
        if row.data:
            return row.data[0]
        return 1.0

    def apply(self, node: NodePath, value: float) -> None:
        """Apply scale to node.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - myNodePath.setScale(S)


        """
        node.setScale(value, value, value)

    def _interpolate(self, last_value: float, next_value: float, factor: float) -> float:
        """Linear interpolation of scale."""
        return last_value + (next_value - last_value) * factor


class Panda3DColorController(Panda3DAnimationController):
    """Color animation controller.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - NodePath.setColor()
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/ColorController.ts:17-110


    """

    def _convert_row_data(self, row: MDLControllerRow) -> Vec3:
        """Convert to RGB Vec3."""
        if len(row.data) >= 3:
            return Vec3(row.data[0], row.data[1], row.data[2])
        return Vec3(1, 1, 1)

    def apply(self, node: NodePath, value: Vec3) -> None:
        """Apply color to node.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - nodePath.setColor(r, g, b, a)


        """
        node.setColor(value.x, value.y, value.z, 1.0)

    def _interpolate(self, last_value: Vec3, next_value: Vec3, factor: float) -> Vec3:
        """Linear interpolation of color."""
        return Vec3(
            last_value.x + (next_value.x - last_value.x) * factor,
            last_value.y + (next_value.y - last_value.y) * factor,
            last_value.z + (next_value.z - last_value.z) * factor,
        )


class Panda3DAlphaController(Panda3DAnimationController):
    """Alpha/transparency animation controller.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - NodePath.setAlphaScale()
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/AlphaController.ts:17-100


    """

    def _convert_row_data(self, row: MDLControllerRow) -> float:
        """Convert to alpha float."""
        if row.data:
            return row.data[0]
        return 1.0

    def apply(self, node: NodePath, value: float) -> None:
        """Apply alpha to node.

        References:
        ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        /panda3d/panda3d-docs - myNodePath.setAlphaScale(SA)


        """
        node.setAlphaScale(value)

    def _interpolate(self, last_value: float, next_value: float, factor: float) -> float:
        """Linear interpolation of alpha."""
        return last_value + (next_value - last_value) * factor


def create_panda3d_controller(mdl_controller: MDLController) -> Panda3DAnimationController | None:
    """Factory function to create appropriate Panda3D controller.

    Args:
    ----
        mdl_controller: MDL controller data from PyKotor

    Returns:
    -------
        Panda3D-specific controller or None if unsupported

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/controllers/OdysseyControllerFactory.ts:36-150
        Libraries/PyKotor/src/pykotor/resource/formats/mdl/mdl_types.py:148-206

    """
    from pykotor.resource.formats.mdl.mdl_types import MDLControllerType

    controller_map = {
        MDLControllerType.Position: Panda3DPositionController,
        MDLControllerType.Orientation: Panda3DOrientationController,
        MDLControllerType.Scale: Panda3DScaleController,
        MDLControllerType.Color: Panda3DColorController,
        MDLControllerType.SelfIllumColor: Panda3DColorController,
        MDLControllerType.Alpha: Panda3DAlphaController,
    }

    controller_class = controller_map.get(mdl_controller.controller_type)
    if controller_class:
        return controller_class(mdl_controller)

    return None


class Panda3DAnimationState(IAnimationState):
    """Panda3D animation state implementation.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Libraries/PyKotor/src/pykotor/engine/animation/base.py:56-92 - IAnimationState
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/OdysseyModelAnimation.ts:25-100


    """

    def __init__(self, animation: MDLAnimation, loop: bool = True, speed: float = 1.0):
        """Initialize animation state."""
        self.animation = animation
        self.current_time: float = 0.0
        self.speed: float = speed
        self.loop: bool = loop
        self.is_playing: bool = False
        self.length: float = animation.length

    def update(self, dt: float) -> bool:
        """Update animation state."""
        if not self.is_playing:
            return False

        self.current_time += dt * self.speed

        if self.current_time >= self.length:
            if self.loop:
                self.current_time = self.current_time % self.length
            else:
                self.current_time = self.length
                self.is_playing = False
                return False

        return True

    def play(self) -> None:
        """Start playing."""
        self.is_playing = True

    def pause(self) -> None:
        """Pause playback."""
        self.is_playing = False

    def stop(self) -> None:
        """Stop and reset."""
        self.is_playing = False
        self.current_time = 0.0


class Panda3DAnimationManager(IAnimationManager):
    """Panda3D animation manager implementation.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Libraries/PyKotor/src/pykotor/engine/animation/base.py:95-181 - IAnimationManager
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/odyssey/OdysseyModelAnimationManager.ts:18-250


    """

    def __init__(self, mdl: MDL, root_node: NodePath):
        """Initialize animation manager."""
        self.mdl = mdl
        self.root_node = root_node
        self.current_animation: Panda3DAnimationState | None = None

        # Build node map
        self.node_map: dict[str, tuple[Any, NodePath]] = {}
        self._build_node_map(mdl.root_node, root_node)

        # Build controllers
        self.animation_controllers: dict[str, dict[str, list[Panda3DAnimationController]]] = {}
        for animation in mdl.animations:
            self._build_controllers_for_animation(animation)

    def _build_node_map(self, mdl_node: Any, node_path: NodePath) -> None:
        """Build map from MDL node names to NodePaths."""
        self.node_map[mdl_node.name] = (mdl_node, node_path)

        for child_mdl, child_np in zip(mdl_node.children, node_path.getChildren()):
            self._build_node_map(child_mdl, child_np)

    def _build_controllers_for_animation(self, animation: MDLAnimation) -> None:
        """Build controllers for animation."""
        animation_name = animation.name
        self.animation_controllers[animation_name] = {}
        self._build_node_controllers(animation.root_node, animation_name)

    def _build_node_controllers(self, mdl_node: Any, animation_name: str) -> None:
        """Build controllers for node recursively."""
        node_controllers: list[Panda3DAnimationController] = []
        for mdl_controller in mdl_node.controllers:
            controller = create_panda3d_controller(mdl_controller)
            if controller:
                node_controllers.append(controller)

        if node_controllers:
            self.animation_controllers[animation_name][mdl_node.name] = node_controllers

        for child in mdl_node.children:
            self._build_node_controllers(child, animation_name)

    def play_animation(self, animation_name: str, loop: bool = True, speed: float = 1.0) -> bool:
        """Play animation by name."""
        animation = None
        for anim in self.mdl.animations:
            if anim.name == animation_name:
                animation = anim
                break

        if not animation:
            return False

        self.current_animation = Panda3DAnimationState(animation, loop, speed)
        self.current_animation.play()

        return True

    def pause_animation(self) -> None:
        """Pause current animation."""
        if self.current_animation:
            self.current_animation.pause()

    def stop_animation(self) -> None:
        """Stop current animation."""
        if self.current_animation:
            self.current_animation.stop()
            self.current_animation = None

    def update(self, dt: float) -> None:
        """Update animation for current frame."""
        if not self.current_animation or not self.current_animation.is_playing:
            return

        still_playing = self.current_animation.update(dt)
        self._apply_animation()

        if not still_playing and not self.current_animation.loop:
            self.current_animation = None

    def _apply_animation(self) -> None:
        """Apply current animation state to nodes."""
        if not self.current_animation:
            return

        animation_name = self.current_animation.animation.name
        current_time = self.current_animation.current_time

        anim_controllers = self.animation_controllers.get(animation_name, {})

        for node_name, controllers in anim_controllers.items():
            if node_name not in self.node_map:
                continue

            mdl_node, node_path = self.node_map[node_name]

            for controller in controllers:
                value = controller.get_value_at_time(current_time)
                if value is not None:
                    controller.apply(node_path, value)

    def get_animation_names(self) -> list[str]:
        """Get all animation names."""
        return [anim.name for anim in self.mdl.animations]

    def has_animation(self, name: str) -> bool:
        """Check if animation exists."""
        return name in self.animation_controllers



"""NCS compilers: inbuilt (native) and external (nwnnsscomp) NSS-to-NCS compilation."""

from __future__ import annotations

import subprocess

from datetime import date
from enum import Enum
from pathlib import Path  # pyright: ignore[reportMissingImports]
from typing import TYPE_CHECKING, NamedTuple

from pykotor.common.misc import Game
from pykotor.resource.formats._base import BiowareResource
from pykotor.resource.formats.ncs.compiler.classes import EntryPointError
from pykotor.resource.formats.ncs.ncs_auto import compile_nss, write_ncs
from pykotor.resource.formats.ncs.ncs_data import NCSCompiler
from pykotor.tools.encoding import decode_bytes_with_fallbacks
from utility.misc import generate_hash  # pyright: ignore[reportMissingImports]

if TYPE_CHECKING:
    import os

    from subprocess import CompletedProcess

    from pykotor.resource.formats.ncs.ncs_data import NCS, NCSOptimizer


class InbuiltNCSCompiler(NCSCompiler):
    """Built-in NSS to NCS compiler using PyKotor's native implementation.

    This compiler provides full NSS compilation without external dependencies,
    supporting all KOTOR/TSL script features including:
    - Functions, variables, structs
    - Control flow (if/else, while, for, do-while, switch)
    - All data types (int, float, string, object, vector, etc.)
    - #include directive support
    - Optimization passes

    References:
    ----------
        See ncs_data module docstring for engine addresses (K1 + TSL TODO). CResNCS::CResNCS (K1: 0x005d4c30),
        ReadScriptFile (K1: 0x005d2260), InitializeScript (K1: 0x005d461b, TSL: TODO). CVirtualMachineInternal — NWScript VM.
        Note: PyKotor NSS compiler produces NCS bytecode compatible with the engine VM.
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/nwscript/NWScriptCompiler.ts (NSS compilation reference)


    """

    def compile_script(  # noqa: PLR0913
        self,
        source_file: os.PathLike | str,
        output_file: os.PathLike | str,
        game: Game | int,
        timeout: int = 5,  # noqa: ARG002
        *,
        debug: bool = False,
        optimizers: list[NCSOptimizer] | None = None,
    ) -> tuple[str, str]:
        if not isinstance(game, Game):
            game = Game(game)
        source_filepath: Path = Path(source_file)
        nss_data: bytes = source_filepath.read_bytes()
        nss_contents: str = decode_bytes_with_fallbacks(nss_data)
        ncs: NCS = compile_nss(nss_contents, game, optimizers, library_lookup=[source_filepath.parent], debug=debug)
        write_ncs(ncs, output_file)
        return "", ""


class ExternalCompilerConfig(NamedTuple):
    sha256: str
    name: str
    release_date: date
    author: str
    commandline: dict[str, list[str]]


class KnownExternalCompilers(Enum):
    """Known external NSS compilers and their configurations.

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/TSLPatcher/tree/master/TSLPatcher.pl (unfinished perl rewrite of TSLPatcher)
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET.Patcher/


    """

    TSLPATCHER = ExternalCompilerConfig(
        sha256="539EB689D2E0D3751AEED273385865278BEF6696C46BC0CAB116B40C3B2FE820",
        name="TSLPatcher",
        release_date=date(2009, 1, 1),
        author="todo",
        commandline={
            "compile": ["-c", "{source}", "-o", "{output}"],
            "decompile": ["-d", "{source}", "-o", "{output}"],
        },
    )
    KOTOR_TOOL = ExternalCompilerConfig(
        sha256="E36AA3172173B654AE20379888EDDC9CF45C62FBEB7AB05061C57B52961C824D",
        name="KOTOR Tool",
        release_date=date(2005, 1, 1),
        author="Fred Tetra",
        commandline={
            "compile": ["-c", "--outputdir", "{output_dir}", "-o", "{output_name}", "-g", "{game_value}", "{source}"],
            "decompile": ["-d", "--outputdir", "{output_dir}", "-o", "{output_name}", "-g", "{game_value}", "{source}"],
        },
    )
    V1 = ExternalCompilerConfig(
        sha256="EC3E657C18A32AD13D28DA0AA3A77911B32D9661EA83CF0D9BCE02E1C4D8499D",
        name="v1.3 first public release",
        release_date=date(2003, 12, 31),
        author="todo",
        commandline={
            "compile": ["-c", "{source}", "{output}"],
            "decompile": ["-d", "{source}", "{output}"],
        },
    )
    KOTOR_SCRIPTING_TOOL = ExternalCompilerConfig(
        sha256="B7344408A47BE8780816CF68F5A171A09640AB47AD1A905B7F87DE30A50A0A92",
        name="KOTOR Scripting Tool",
        release_date=date(2016, 5, 18),
        author="James Goad",  # TODO: double check
        commandline={
            "compile": ["-c", "--outputdir", "{output_dir}", "-o", "{output_name}", "-g", "{game_value}", "{source}"],
            "decompile": ["-d", "--outputdir", "{output_dir}", "-o", "{output_name}", "-g", "{game_value}", "{source}"],
        },
    )
    DENCS = ExternalCompilerConfig(
        sha256="539EB689D2E0D3751AEED273385865278BEF6696C46BC0CAB116B40C3B2FE820",
        name="DeNCS",
        release_date=date(2006, 5, 30),
        author="todo",
        commandline={},
    )
    XOREOS = ExternalCompilerConfig(
        sha256="",
        name="Xoreos Tools",
        release_date=date(2016, 1, 1),  # Approximate based on project history
        author="Xoreos Team",
        commandline={},  # Xoreos tools are primarily for engine reimplementation
    )
    KNSSCOMP = ExternalCompilerConfig(
        # knsscomp is Nick Hugi's modern NSS compiler
        # Hash should be determined per binary version
        sha256="",  # TODO: Obtain hash from actual knsscomp binary
        name="knsscomp",
        release_date=date(2022, 1, 1),  # Approximate
        author="Nick Hugi",
        commandline={
            "compile": ["-c", "{source}", "-o", "{output}"],
            "decompile": [],  # knsscomp doesn't support decompilation
        },
    )

    @classmethod
    def from_sha256(cls: type[KnownExternalCompilers], sha256: str) -> KnownExternalCompilers:
        uppercase_sha256: str = sha256.upper()
        for known_ext_compiler in cls:
            if known_ext_compiler.value.sha256 == uppercase_sha256:
                return known_ext_compiler

        msg = f"No compilers found with sha256 hash '{uppercase_sha256}'"
        raise ValueError(msg)


class NwnnsscompConfig(BiowareResource):
    """Unifies the arguments passed to each different version of nwnnsscomp, since no versions offer backwards-compatibility with each other."""

    def __init__(
        self,
        sha256_hash: str,
        sourcefile: Path,
        outputfile: Path,
        game: Game,
    ):
        self.sha256_hash: str = sha256_hash
        self.source_file: Path = sourcefile
        self.output_file: Path = outputfile
        self.output_dir: Path = outputfile.parent
        self.output_name: str = outputfile.name
        self.game: Game = game

        self.chosen_compiler: KnownExternalCompilers = KnownExternalCompilers.from_sha256(self.sha256_hash)

    def get_compile_args(self, executable: str) -> list[str]:
        return self._format_args(self.chosen_compiler.value.commandline["compile"], executable)

    def get_decompile_args(self, executable: str) -> list[str]:
        return self._format_args(self.chosen_compiler.value.commandline["decompile"], executable)

    def _format_args(self, args_list: list[str], executable: str) -> list[str]:
        formatted_args: list[str] = [
            arg.format(
                source=self.source_file,
                output=self.output_file,
                output_dir=self.output_dir,
                output_name=self.output_name,
                game_value="1" if self.game.is_k1() else "2",
            )
            for arg in args_list
        ]
        formatted_args.insert(0, executable)
        return formatted_args


class ExternalNCSCompiler(NCSCompiler):
    def __init__(self, nwnnsscomp_path: os.PathLike | str):
        self.change_nwnnsscomp_path(nwnnsscomp_path)

    def get_info(self) -> KnownExternalCompilers:
        return KnownExternalCompilers.from_sha256(self.filehash)

    def change_nwnnsscomp_path(self, nwnnsscomp_path: os.PathLike | str):
        self.nwnnsscomp_path: Path = Path(nwnnsscomp_path)
        self.filehash: str = generate_hash(self.nwnnsscomp_path, hash_algo="sha256").upper()

    def config(
        self,
        source_file: os.PathLike | str,
        output_file: os.PathLike | str,
        game: Game | int,
        *,
        debug: bool = False,
    ) -> NwnnsscompConfig:
        """Configures a Nwnnsscomp run.

        Args:
        ----
            source_file: Path to the source file to compile
            output_file: Path to output file to generate
            game: Game enum or integer to configure in one line
            debug - bool (kwarg): Whether to debug Ply and verbosely output more information. Defaults to False.

        Returns:
        -------
            NwnnsscompConfig: Config object for Nwnnsscomp run

        Processing Logic:
        ----------------
            - Resolves source and output file paths
            - Converts game arg to Game enum if integer
            - Returns NwnnsscompConfig object configured with args useable with the compile_script and decompile_script functions.
        """
        source_filepath, output_filepath = map(Path, (source_file, output_file))
        if not isinstance(game, Game):
            game = Game(game)
        return NwnnsscompConfig(self.filehash, source_filepath, output_filepath, game)

    def compile_script(
        self,
        source_file: os.PathLike | str,
        output_file: os.PathLike | str,
        game: Game | int,
        timeout: int = 5,
        *,
        debug: bool = False,
    ) -> tuple[str, str]:
        """Compiles a NSS script into NCS using the external compiler.

        Function is compatible with any nwnnsscomp.exe version.

        Args:
        ----
            source_file: The path or name of the script source file to compile.
            output_file: The path or name of the compiled module file to output.
            game: The Game object or game ID to configure the compiler for.
            timeout: The timeout in seconds to wait for compilation to finish before aborting.
            debug - bool (kwarg): (does nothing for external compilers)

        Returns:
        -------
            A tuple of (stdout, stderr) strings from the compilation process.

        Processing Logic:
        ----------------
            - Validates source file exists
            - Configures the compiler based on the nwnnsscomp.exe used
            - Runs the compiler process, capturing stdout and stderr
            - Returns a tuple of the stdout and stderr strings on completion

        Raises:
        ------
            FileNotFoundError: If source file doesn't exist
            RuntimeError: If compiler executable doesn't exist
            EntryPointError: If file has no entry point and is an include file
            subprocess.TimeoutExpired: If compilation exceeds timeout
        """
        source_path = Path(source_file)
        if not source_path.is_file():
            msg = f"Source file not found: {source_path}"
            raise FileNotFoundError(msg)

        if not self.nwnnsscomp_path.is_file():
            msg = f"Compiler executable not found: {self.nwnnsscomp_path}"
            raise RuntimeError(msg)

        config: NwnnsscompConfig = self.config(source_file, output_file, game)

        try:
            result: CompletedProcess[str] = subprocess.run(
                args=config.get_compile_args(str(self.nwnnsscomp_path)),
                capture_output=True,  # Capture stdout and stderr
                text=True,
                timeout=timeout,
                check=False,
            )
        except subprocess.TimeoutExpired as e:
            msg = f"Compilation timed out after {timeout} seconds"
            raise RuntimeError(msg) from e
        except Exception as e:
            msg = f"Failed to run compiler: {e}"
            raise RuntimeError(msg) from e

        stdout, stderr = self._get_output(result)

        # Check for known error conditions
        if "File is an include file, ignored" in stdout:
            msg = "This file has no entry point and cannot be compiled (Most likely an include file)."
            raise EntryPointError(msg)

        if result.returncode != 0 and stderr:
            msg = f"Compilation failed with return code {result.returncode}: {stderr}"
            raise RuntimeError(msg)

        return stdout, stderr

    def decompile_script(
        self,
        source_file: os.PathLike | str,
        output_file: os.PathLike | str,
        game: Game | int,
        timeout: int = 5,
    ) -> tuple[str, str]:
        """Decompiles a NCS bytecode file to NSS source code.

        Args:
        ----
            source_file: Path to the NCS bytecode file to decompile
            output_file: Path to output the decompiled NSS source code
            game: The Game object or ID for configuration
            timeout: Seconds to wait before aborting (default: 5)

        Returns:
        -------
            tuple[str, str]: Tuple of (stdout, stderr) from the decompile process

        Processing Logic:
        ----------------
            - Validates source file exists
            - Validates compiler supports decompilation
            - Runs the decompile process and captures output
            - Returns stdout/stderr streams

        Raises:
        ------
            FileNotFoundError: If source file doesn't exist
            RuntimeError: If compiler doesn't support decompilation or fails
            subprocess.TimeoutExpired: If decompilation exceeds timeout
        """
        source_path = Path(source_file)
        if not source_path.is_file():
            msg = f"Source file not found: {source_path}"
            raise FileNotFoundError(msg)

        if not self.nwnnsscomp_path.is_file():
            msg = f"Compiler executable not found: {self.nwnnsscomp_path}"
            raise RuntimeError(msg)

        config: NwnnsscompConfig = self.config(source_file, output_file, game)

        # Check if this compiler supports decompilation
        if not config.chosen_compiler.value.commandline.get("decompile"):
            msg = f"Compiler '{config.chosen_compiler.value.name}' does not support decompilation"
            raise RuntimeError(msg)

        try:
            result: CompletedProcess[str] = subprocess.run(
                args=config.get_decompile_args(str(self.nwnnsscomp_path)),
                capture_output=True,  # Capture stdout and stderr
                text=True,
                timeout=timeout,
                check=False,
            )
        except subprocess.TimeoutExpired as e:
            msg = f"Decompilation timed out after {timeout} seconds"
            raise RuntimeError(msg) from e
        except Exception as e:
            msg = f"Failed to run decompiler: {e}"
            raise RuntimeError(msg) from e

        stdout, stderr = self._get_output(result)

        if result.returncode != 0 and stderr:
            msg = f"Decompilation failed with return code {result.returncode}: {stderr}"
            raise RuntimeError(msg)

        return stdout, stderr

    def _get_output(self, result: CompletedProcess[str]) -> tuple[str, str]:
        stdout: str = result.stdout
        stderr: str = (
            f"No error provided, but return code is nonzero: ({result.returncode})"
            if result.returncode != 0 and (not result.stderr or not result.stderr.strip())
            else result.stderr
        )

        if "Error:" in stdout:
            stdout_lines: list[str] = stdout.split("\n")
            error_line: str = ""
            # Find and remove the line with 'Error:'
            filtered_stdout_lines: list[str] = []
            for line in stdout_lines:
                if "Error:" in line:
                    error_line += "\n" + line
                else:
                    filtered_stdout_lines.append(line)

            # Reconstruct stdout without the error line
            stdout = "\n".join(filtered_stdout_lines)

            # Append the error line to stderr if it was found
            if error_line:
                if stderr:  # If there's already content in stderr, add a newline before appending
                    stderr += "\n" + error_line
                else:
                    stderr = error_line
        return stdout, stderr
